# -*- coding: utf-8 -*-
"""cevlib"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")


from typing import List
import re

import aiohttp

from cevlib.types.iType import IType, JObject


class Featured(IType):
    """featured news, images, ..."""
    def __init__(self) -> None:
        self._gallery: List[str] = [ ]
        self._videos: List[str] = [ ]

    async def init(self) -> None:
        """init"""
        async with aiohttp.ClientSession() as client:
            async with client.get("https://www.cev.eu/") as resp:
                html = await resp.text()
        self._gallery = [ f"https://www.cev.eu{match[0]}"
                          for match in re.finditer(r"(\/media\/[\w .,@;?^=%&:\/~+#-]*[\w@?^=%&\/~+#-]).(jpg|JPG)", html) ] # pylint: disable=line-too-long
        self._videos = [ f"https://{match[0].replace('/embed/', '/v/').split('?')[0]}"
                         for match in re.finditer(r"([\w_-]+(?:(?:\.[\w_-]+)+))([\w.,@;?^=%&:\/~+#-]*\/embed\/[\w .,@;?^=%&:\/~+#-]*[\w@?^=%&\/~+#-])", html) ] # pylint: disable=line-too-long

    @property
    def gallery(self) -> List[str]:
        """gallery"""
        return self._gallery

    @property
    def videos(self) -> List[str]:
        """videos"""
        return self._videos

    def toJson(self) -> JObject:
        return {
            "gallery": self.gallery,
            "videos": self.videos
        }

    @property
    def valid(self) -> bool:
        return True



# -*- coding: utf-8 -*-
"""cevlib"""
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from typing import Any, Callable, Optional, TypeVar, cast, Iterable
from threading import Thread
import asyncio
from queue import Queue

T = TypeVar("T")


async def asyncRunInThread(target: Callable[[Any], None],
                           *args: Optional[Iterable[Any]]) -> None:
    """runs the callable in a thread while providing an async interface for it"""
    thread = Thread(target = target, args = args)
    thread.start()
    while thread.is_alive():
        await asyncio.sleep(1)

async def asyncRunInThreadWithReturn(target: Callable[[Any], T],
                                     *args: Optional[Iterable[Any]]) -> T:
    """
    runs the callable in a thread while providing an async interface for it
    (allows return value)
    """
    queue: Queue[Any] = Queue()

    def _implement() -> None:
        ret = target(*args) if args else target() # type: ignore
        queue.put_nowait(ret)

    thread = Thread(target = _implement)
    thread.start()
    while thread.is_alive():
        await asyncio.sleep(1)
    return cast(T, queue.get_nowait())



# -*- coding: utf-8 -*-
"""cevlib"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from typing import List

from cevlib.helpers.dictTool import DictEx, ListEx

from cevlib.types.iType import IType, JObject
from cevlib.types.results import SetResult
from cevlib.types.types import PlayType


class Play(IType):
    """play/event"""
    def __init__(self, data: JObject) -> None:
        dex = DictEx(data)
        self._type: PlayType = PlayType.parse(dex.assertGet("Title", str))
        self._currentScore: SetResult = SetResult.parseFromPlayByPlay(data)
        self._playerName: str = dex.assertGet("PlayerName", str).title() # TODO player type
        self._playerNumber: int = dex.assertGet("PlayerNumber", int)
        self._isHome: bool = dex.assertGet("IsHome", bool)

    def toJson(self) -> JObject:
        return {
            "type": self.type.value,
            "currentScore": self.currentScore.toJson(),
            "playerName": self.playerName,
            "playerNumber": self._playerNumber,
            "isHome": self._isHome
        }

    @property
    def valid(self) -> bool:
        return None not in (self._type, self._currentScore, self._playerName)

    @property
    def type(self) -> PlayType:
        """play type (e.g. block, spike, ...)"""
        return self._type

    @property
    def currentScore(self) -> SetResult: # TODO rename to result?
        """current score at time of play"""
        return self._currentScore

    @property
    def playerName(self) -> str:
        """player that performed that play"""
        return self._playerName

    def __repr__(self) -> str:
        return f"(cevlib.types.playByPlay.Play) {self._type} {self._currentScore} by {self._playerName}" # pylint: disable=line-too-long


class Set(IType):
    """play by play set"""
    def __init__(self, data: JObject) -> None:
        dex = DictEx(data)
        self._plays = [ Play(event) for event in dex.ensure("Events", list) ]
        self._setNumber = ListEx(dex.ensure("TabName", str).split(" ")).ensure(1, int)

    def toJson(self) -> JObject:
        return {
            "plays": [ play.toJson() for play in self.plays ],
            "setNumber": self.setNumber
        }

    @property
    def valid(self) -> bool:
        return bool(self._plays)

    @property
    def plays(self) -> List[Play]:
        """all plays/events"""
        return self._plays

    @property
    def setNumber(self) -> int:
        """set number (1 - 5)"""
        return self._setNumber

    def __repr__(self) -> str:
        return f"(cevlib.types.playByPlay.Set) Set {self._setNumber} {self._plays}"


class PlayByPlay(IType):
    """all plays/events across all sets"""
    def __init__(self, data: JObject) -> None:
        self._sets = [ Set(playEvent)
                       for playEvent in DictEx(data).ensure("PlayEvents", list) ]

    @property
    def valid(self) -> bool:
        return bool(self._sets)

    @property
    def sets(self) -> List[Set]:
        """all sets"""
        return self._sets

    def toJson(self) -> JObject:
        return {
            "sets": [ set_.toJson() for set_ in self.sets ]
        }

    def __repr__(self) -> str:
        return f"(cevlib.types.playByPlay.PlayByPlay) {self._sets}"



# -*- coding: utf-8 -*-
"""cevlib"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from typing import Any, Dict, List, Optional, Tuple
import re
from bs4 import BeautifulSoup # type: ignore
from bs4.element import Tag # type: ignore

import aiohttp

from cevlib.calendar import CalendarMatch

from cevlib.helpers.dictTool import DictEx

from cevlib.types.competition import MatchCompetition
from cevlib.types.iType import IType
from cevlib.types.results import Result
from cevlib.types.team import Team
from cevlib.types.types import CompetitionGender


class Draw(IType):
    """competition draw. consists of two calendarMatches"""
    def __init__(self, matches: List[CalendarMatch]) -> None:
        self._matches = matches

        assert len(matches) <= 2
        self._first, self._second = matches[0].teams

    def _sameTeams(self) -> bool:
        for match in self._matches:
            if self._first not in match.teams:
                return False
            if self._second not in match.teams:
                return False
        return True

    @property
    def competition(self) -> MatchCompetition:
        """competition"""
        return self._matches[0].competition

    @property
    def firstTeam(self) -> Team:
        """first team"""
        return self._first

    @property
    def secondTeam(self) -> Team:
        """second team"""
        return self._second

    @property
    def teams(self) -> Tuple[Team, Team]:
        """both teams"""
        assert self.valid
        return self._matches[0].teams

    @property
    def matches(self) -> List[CalendarMatch]:
        """both matches of the draw"""
        assert self.valid
        return self._matches

    def get(self, index: int) -> Optional[CalendarMatch]:
        """matches[index]"""
        return self.matches[index] if index < len(self.matches) else None

    @property
    def valid(self) -> bool:
        return bool(self._matches) and self._sameTeams()

    def __repr__(self) -> str:
        return f"(cevlib.competitions.Draw) {len(self._matches)} matches"

    def toJson(self) -> List[Dict[str, Any]]:
        if not self.valid:
            print(self._first, self._second, self._sameTeams())
        return [ match.toJson() for match in self._matches ]


class StandingsTeam(IType):
    """standings team"""
    def __init__(self, team: Dict[str, Any]) -> None:
        self._team = team

    @property
    def team(self) -> Dict[str, Any]:
        """the team"""
        return self._team

    @property
    def valid(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"(cevlib.competitions.StandingsTeam) {self.team}"

    def toJson(self) -> Dict[str, Any]:
        return self._team


class StandingsPool(IType):
    """standings pool (i.e. group). consists of multiple standingsTeams"""
    def __init__(self, teams: List[StandingsTeam]) -> None:
        self._teams = teams

    @property
    def teams(self) -> List[StandingsTeam]:
        """all teams"""
        return self._teams

    @property
    def valid(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"(cevlib.competitions.StandingsPool) {len(self.teams)} teams"

    def toJson(self) -> List[Dict[str, Any]]:
        return [ team.toJson() for team in self.teams ]

class Standings(IType):
    """standings (e.g. score tables). consists of multiple standingsPools"""
    def __init__(self, table: Optional[Tag]) -> None:
        self._pools: List[StandingsPool] = [ ]
        if not table:
            return
        head = table.find("thead")
        bodies = table.find_all("tbody")
        if not head or len(bodies) == 0:
            return
        rows = head.find_all("tr")
        if len(rows) < 2:
            return
        headings1 = [ th.get_text(strip = True) for th in rows[0].find_all("th") ]
        headings1 = [ heading for heading in headings1 if heading ]
        headings1.insert(0, "")
        self._headings = [ ]
        headings1It = 0
        for th in rows[1].find_all("th"):
            assert isinstance(th, Tag)
            self._headings.append(f"{headings1[headings1It]}.{th.get_text(strip = True)}")
            if th.get("class") == ["u-pr-8"]:
                headings1It += 1

        if len(self._headings) == 0:
            return
        for body in bodies:
            teams: List[StandingsTeam] = [ ]
            for row in body.find_all("tr"):
                standing = { }
                for i, text in enumerate(row.find_all("td")):
                    standing[self._headings[i]] = text.get_text(strip = True)
                teams.append(StandingsTeam(standing))
            self._pools.append(StandingsPool(teams))

    @property
    def pools(self) -> List[StandingsPool]:
        """standings pools"""
        return self._pools

    def get(self, index: int) -> Optional[StandingsPool]:
        """pools[index]"""
        if index >= len(self._pools):
            return None
        return self._pools[index]

    @property
    def valid(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"(cevlib.competitions.Standings) {len(self.pools)} pools"

    def toJson(self) -> List[List[Dict[str, Any]]]:
        return [ pool.toJson() for pool in self.pools ]


class Pool(IType):
    """round pool. consists of either standings or draws"""
    def __init__(self,
                 name: str,
                 draws: List[Draw],
                 standings: Optional[StandingsPool] = None) -> None:
        self._name = name
        self._draws = draws
        self._standings = standings

    def moveOrCreateDraw(self,
                         anyDrawTeam: Team,
                         newIndex: int,
                         competition: Optional[MatchCompetition]) -> None:
        """arranges the draw to match the right node. if the draw is not found, it is created"""
        if anyDrawTeam.name == "Bye":
            return
        oldIndex = -1
        for i, _ in enumerate(self._draws):
            if anyDrawTeam in self._draws[i].teams:
                oldIndex = i
                break

        if oldIndex >= 0:
            self._draws.insert(newIndex, self._draws.pop(oldIndex))
            return

        assert competition
        self._draws.insert(newIndex,
                           Draw([ CalendarMatch.shortcutMatch(competition, anyDrawTeam),
                                  CalendarMatch.shortcutMatch(competition, anyDrawTeam) ]))

    @property
    def valid(self) -> bool:
        return bool(self._name and len(self._draws))

    @property
    def standingsPool(self) -> Optional[StandingsPool]:
        """pool (i.e. group) standings"""
        return self._standings

    @property
    def draws(self) -> List[Draw]:
        """pool draws"""
        return self._draws

    def get(self, index: int) -> Optional[Draw]:
        """draws[index]"""
        return self.draws[index] if index < len(self.draws) else None

    @property
    def name(self) -> str:
        """pool name"""
        return self._name

    def __repr__(self) -> str:
        return f"(cevlib.competitions.Pool) {self._name} ({len(self.draws)} draws) ({self.standingsPool})" # pylint: disable=line-too-long

    def toJson(self) -> Dict[str, Any]:
        return {
            "name": self._name,
            "draws": [ draw.toJson() for draw in self._draws ],
            "standings": self.standingsPool.toJson() if self.standingsPool else None
        }


class Round(IType):
    """competition round. consists of multiple pools."""
    def __init__(self, name: str, pools: List[Pool]) -> None:
        self._name = name
        self._pools = pools

    @property
    def valid(self) -> bool:
        return bool(self._name and len(self._pools))

    @property
    def name(self) -> str:
        """round name"""
        return self._name

    @property
    def pools(self) -> List[Pool]:
        """round pools"""
        return self._pools

    def get(self, index: int) -> Optional[Pool]:
        """pools[index]"""
        return self.pools[index] if index < len(self.pools) else None

    @staticmethod
    def parsePool(pool: Dict[str, Any],
                  competition: CompetitionLink,
                  standings: Optional[StandingsPool]) -> Pool:
        """parse pool from json"""
        pdex = DictEx(pool)
        draws: List[List[CalendarMatch]] = [ ]
        for match in pdex.ensureList("Results"):
            mdex = DictEx(match)
            homeId = 0
            awayId = 0
            try: # TODO safer regex solution
                homeId = int(re.search(r"^\/team\/([0-9]+)-[\w-]+$", match["HomeTeam"]["Link"]).group(1)) # type: ignore # pylint: disable=line-too-long
                awayId = int(re.search(r"^\/team\/([0-9]+)-[\w-]+$", match["AwayTeam"]["Link"]).group(1)) # type: ignore # pylint: disable=line-too-long
            except Exception as exc:
                print(exc)

            newMatch = CalendarMatch(mdex.ensureString("MatchCentreUrl"),
                MatchCompetition.buildPlain(competition.name,
                    competition.gender,
                    season = str(competition.age) if competition.age else None,
                    phase = pdex.ensure("Name", str),
                    matchNumber = mdex.ensure("MatchName", str)),
                Team.build(mdex.ensure("HomeTeam", DictEx).ensure("Name", str),
                    mdex.ensure("HomeTeam", DictEx).ensure("Logo", DictEx).ensure("Name", str),
                    "N/A",
                    True,
                    homeId),
                Team.build(mdex.ensure("AwayTeam", DictEx).ensure("Name", str),
                    mdex.ensure("AwayTeam", DictEx).ensure("Logo", DictEx).ensure("Name", str),
                    "N/A",
                    False,
                    awayId),
                mdex.ensureString("Location"),
                mdex.ensureString("MatchDateTime"),
                Result.parseFromForm(match),
                mdex.ensureBool("IsComplete"))
            newDraw = True
            for draw in draws:
                if draw[0].awayTeam.name == newMatch.homeTeam.name and \
                   draw[0].homeTeam.name == newMatch.awayTeam.name:
                    newDraw = False
                    draw.append(newMatch)
                    break
            if newDraw:
                draws.append([newMatch])
        return Pool(pdex.ensureString("Name"),
                    [ Draw(draw)
                      for draw in draws ],
                    standings)

    def __repr__(self) -> str:
        return f"(cevlib.competitions.Round) {self._name} ({len(self.pools)} pools)"

    def toJson(self) -> Dict[str, Any]:
        return {
            "name": self._name,
            "pools": [ pool.toJson() for pool in self._pools ]
        }


class Competition(IType):
    """a single competition. consists of multiple rounds"""
    def __init__(self, rounds: List[Round]) -> None:
        self._rounds = rounds
        self._rearrangeDraws()

    def _rearrangeDraws(self) -> None:
        pools: List[Pool] = [ ]
        for round_ in self._rounds:
            pools.extend(round_.pools)
        for i in range(len(pools) - 1):
            pool = pools[- (i + 1)]
            previousPool = pools[- (i + 2)]
            if pool.standingsPool or previousPool.standingsPool:
                continue
            for j, _ in enumerate(pool.draws):
                previousPool.moveOrCreateDraw(pool.draws[j].firstTeam,
                                              j * 2, pool.draws[j].competition)
                previousPool.moveOrCreateDraw(pool.draws[j].secondTeam,
                                              j * 2 + 1, pool.draws[j].competition)

    @staticmethod
    def _parseRound(name: str,
                    pools: List[Dict[str, Any]],
                    competition: CompetitionLink,
                    standings: Optional[Standings] = None) -> Round:
        return Round(name,
                    [ Round.parsePool(pool,
                                      competition,
                                      standings.get(i) if standings else None)
                      for i, pool in enumerate(pools) ])

    @staticmethod
    async def fromUrl(url: str) -> Competition:
        """parse a competition from a url"""
        html = ""
        competition = (await Competitions.getAll()).getByLink(url)
        assert competition
        async with aiohttp.ClientSession() as client:
            async with client.get(url) as resp:
                html = await resp.text()
        rounds = [ ]
        soup = BeautifulSoup(html, "html.parser")

        roundNameLookup = [ comp.get_text(strip = True)
                            for comp in soup.find_all("li", class_="tabs-title") ]

        #data-score-endpoint
        for i, comp in enumerate(soup.find_all("div", class_="competition-components-container")):
            if not isinstance(comp, Tag):
                continue
            linkDiv = comp.find("div", attrs = { "data-score-endpoint": True })
            if not isinstance(linkDiv, Tag):
                continue # has no useful content
            tableDiv = comp.find("div", class_="pool-standings-table")
            standings = Standings(tableDiv)
            link = "https:" + linkDiv["data-score-endpoint"]
            async with aiohttp.ClientSession() as client:
                async with client.get(link) as resp:
                    jdata = await resp.json(content_type=None)
                    name = roundNameLookup[i] if i <= len(roundNameLookup) else "N/A"
                    rounds.append(Competition._parseRound(name,
                                                          jdata.get("Pools"),
                                                          competition,
                                                          standings))
        return Competition(rounds)

    @property
    def rounds(self) -> List[Round]:
        """returns the rounds of the competition"""
        return self._rounds

    def get(self, index: int) -> Optional[Round]:
        """rounds[index]"""
        return self.rounds[index] if index < len(self.rounds) else None

    @property
    def valid(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"(cevlib.competitions.Competition) {len(self._rounds)} rounds"

    def toJson(self) -> List[Dict[str, Any]]:
        return [ round.toJson() for round in self._rounds ]


class CompetitionLink(IType):
    """minimal information about a competition"""
    def __init__(self, menuTitle: str, slabTitle: str, entry: str, href: str) -> None:
        self._href = href
        self._type = menuTitle
        self._name = slabTitle
        self._gender = CompetitionGender.parse(entry)
        self._age: Optional[int] = None

        if self._gender == CompetitionGender.Unknown:
            match = re.search(r'U([0-9]{2})([MW])', entry)
            assert match # TODO safer regex search solution
            age = match.group(1)
            gender = match.group(2)
            self._age = int(age) if age else None
            self._gender = CompetitionGender.parse(gender)

    @property
    def name(self) -> str:
        """competition name"""
        return self._name

    @property
    def type(self) -> str:
        """competition type"""
        return self._type

    @property
    def href(self) -> str:
        """competition href"""
        return self._href

    @property
    def gender(self) -> CompetitionGender:
        """gender of the players"""
        return self._gender

    @property
    def age(self) -> Optional[int]:
        """max player age"""
        return self._age

    @property
    def valid(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"(cevlib.competitions.CompetitionLink) {self.name} ({self.type}/{self.age}) {self._gender} ({self.href})" # pylint: disable=line-too-long

    def toJson(self) -> Dict[str, Any]:
        return {
            "name": self._name,
            "type": self._type,
            "gender": self._gender.value,
            "maxAge": self._age,
            "href": self._href
        }

    async def toCompetition(self) -> Competition:
        """cast to full competition"""
        return await Competition.fromUrl(self.href)


class Competitions(IType):
    """competitions wrapper (create with .getAll())"""
    _competitionsCache: List[CompetitionLink] = [ ]

    def __init__(self, html: str) -> None:
        self._competitions: List[CompetitionLink] = [ ]
        if len(Competitions._competitionsCache) > 0:
            self._competitions = Competitions._competitionsCache
            return
        soup = BeautifulSoup(html, "html.parser")
        for menuItem in soup.find_all("li", class_="c-nav__list__item"):
            if not isinstance(menuItem, Tag):
                continue
            menuTitleEl = menuItem.find("a", class_="menuItem")
            if not menuTitleEl:
                continue
            menuTitle = menuTitleEl.get_text(strip = True)

            for slab in menuItem.find_all("div", class_="menuSlab"):
                if not isinstance(slab, Tag):
                    continue
                for row in slab.find_all("div", class_="menuSlab__row"):
                    if not isinstance(row, Tag):
                        continue

                    for rowItem in row.find_all("div"):
                        titleEl = rowItem.find("a", class_="title")
                        if not titleEl:
                            continue

                        title = titleEl.get_text()

                        for item in rowItem.find_all("li"):
                            if not isinstance(item, Tag):
                                continue
                            atag = item.find("a")
                            if not atag:
                                continue
                            itemTitle = atag.attrs.get("title")
                            href = atag.attrs.get("href")

                            string = f"{menuTitle} - {title} - {itemTitle} ({href})"
                            canAppend = True
                            for block in [ "ranking", # TODO beach, snow, national
                                           "history",
                                           "multi-sport events",
                                           "latest",
                                           "beach",
                                           "snow",
                                           "nationals" ]:
                                if block in string.lower():
                                    canAppend = False
                                    continue
                            if canAppend:
                                self._competitions.append(CompetitionLink(menuTitle,
                                                                          title,
                                                                          itemTitle,
                                                                          href))

    @staticmethod
    async def getAll() -> Competitions:
        """get all competitions"""
        async with aiohttp.ClientSession() as client:
            async with client.get("https://www.cev.eu/") as resp:
                return Competitions(await resp.text())

    @property
    def valid(self) -> bool:
        return True

    def getByLink(self, link: str) -> Optional[CompetitionLink]:
        """get competitionLink (basic info) by link"""
        return next((item for item in self._competitions if item.href == link), None)

    @property
    def links(self) -> List[CompetitionLink]:
        """returns all competition links"""
        return self._competitions

    def get(self, index: int) -> Optional[CompetitionLink]:
        """links[index]"""
        return self.links[index] if index < len(self.links) else None

    def __repr__(self) -> str:
        return f"(cevlib.competitions.Competitions) {self._competitions}"

    def toJson(self) -> List[Dict[str, Any]]:
        return [ competition.toJson() for competition in self._competitions ]



import asyncio

from cevlib.match import Match

async def main() -> None:
    # get a match object by its match centre url
    match: Match = await Match.byUrl("https://championsleague.cev.eu/en/match-centres/cev-champions-league-volley-2022/men/clm-61-cucine-lube-civitanova-v-ok-merkur-maribor/")

    # print all available properties

    print(await match.playByPlay())
    print(await match.competition())
    print(await match.topPlayers())
    print(await match.report())
    print(match.gallery) # Disclaimer: Photos featured on the CEV Photo Galleries are downloadable copyright free for media purposes only and only if CEV is credited as the source material. They are protected by copyright for all other commercial purposes. Those wishing to use CEV Photo Gallery photos for other commercial purposes should contact press@cev.eu
    print(match.matchCentreLink)

    await match.init()

    print(await match.result())
    print(await match.duration())
    print(await match.startTime())
    print(await match.venue())
    print(await match.homeTeam())
    print(await match.watchLink())
    print(await match.highlightsLink())
    print(await match.state())

asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
asyncio.run(main())



import asyncio
from cevlib.competitions import Competitions, Competition

async def main() -> None:
    competitions = await Competitions.getAll() # get all competitions (Competitions object)
    print( competitions )

    link = competitions.get(0) # get the first competition (CompetitionLink object)
    print( link )

    assert link
    competition = await link.toCompetition()
    print( competition )

    # or
    competition = await Competition.fromUrl(link.href) # get the first competition (Competition object)
    print( competition )

    assert competition
    round = competition.get(1) # get the first round (Round object)
    print( round )

    assert round
    pool = round.get(0) # get the first pool (Pool object)
    print( pool )

    assert pool
    standingsPool = pool.standingsPool # get this pool's standings (StandingsPool object)
    print( standingsPool )

    assert pool
    draw = pool.get(0) # get the first draw (Draw object)
    print( draw )

    assert draw
    match = draw.get(0) # get the first match (CalendarMatch object)
    print ( match )

asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
asyncio.run(main())



# -*- coding: utf-8 -*-
"""cevlib"""
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from cevlib.helpers.dictTool import DictEx, ListEx

from cevlib.types.iType import JArray, JObject


class ScoreHeroToJson:
    """score hero endpoint to live scores endpoint data"""
    @staticmethod
    def convert(data: JObject, matchPollData: JArray) -> JObject:
        """convert"""
        dex = DictEx(data)
        pollData = ListEx(matchPollData)
        return {
            "matchLocation": dex.tryGet("StadiumInformation", str),
            "utcStartDate": dex.tryGet("MatchStartDateTimeUTC", str),
            "matchId": dex.ensure("MatchId", int),
            "homeSetsWon": dex.ensure("HomeTeam", DictEx).tryGet("Score", int),
            "awaySetsWon": dex.ensure("AwayTeam", DictEx).tryGet("Score", int),
            "hasGoldenSet": dex.ensure("GoldenSet", bool),
            "setResults": ScoreHeroToJson._convertSets(data),
            "matchState_String": "FINISHED",
            "matchNumber": "MatchNumber",
            "currentSetScore": {
                "homeScore": 0,
                "awayScore": 0,
                "setNumber": 0,
                "isInPlay": False
            },
            "homeTeam": dex.ensure("HomeTeam", DictEx).tryGet("Name", str),
            "awayTeam": dex.ensure("AwayTeam", DictEx).tryGet("Name", str),
            "homeTeamIcon": dex.ensure("HomeTeam", DictEx)
                               .ensure("Logo", DictEx).tryGet("Url", str),
            "awayTeamIcon": dex.ensure("AwayTeam", DictEx)
                               .ensure("Logo", DictEx).tryGet("Url", str),
            "homeTeamNickname": pollData.ensure(0, DictEx).tryGet("Value", str),
            "awayTeamNickname": pollData.ensure(1, DictEx).tryGet("Value", str),
            "homeTeamId": pollData.ensure(0, DictEx).tryGet("Id", int),
            "awayTeamId": pollData.ensure(1, DictEx).tryGet("Id", int)
        }

    @staticmethod
    def _convertSets(data: JObject) -> JArray:
        """convert sets"""
        mainSets = data["SetsFormatted"]\
                        .replace("<span>", "")\
                        .replace("</span>", "")\
                        .replace("(", "")\
                        .replace(")", "")\
                        .replace(" ", "")\
                        .split(",")
        goldenSet = data["GoldenSet"]
        sets = [ ]
        index = 1
        for mainSet in mainSets:
            if mainSet:
                sets.append({
                    "homeScore": mainSet.split("-")[0],
                    "awayScore": mainSet.split("-")[1],
                    "setNumber": index,
                    "isInPlay": False
                })
                index += 1
        if goldenSet:
            sets.append({
                "homeScore": goldenSet.split("-")[0],
                "awayScore": goldenSet.split("-")[1],
                "setNumber": index,
                "isInPlay": False
            })
        return sets



import asyncio
from cevlib.calendar import Calendar

async def main() -> None:
    matches = await Calendar.matchesOfMonth(3, 2022) # get all matches of march, 2022
    print( matches )

    matches = await Calendar.upcomingMatches() # get approximately 10 upcoming (or running) games
    # these matches are displayed on cev.eu under "GameHub/Match List"
    print( matches )

    matches = await Calendar.recentMatches() # get approximately 10 recently finished games
    # these matches are displayed on cev.eu under "GameHub/Recent Results"
    print( matches )

    matches = await Calendar.upcomingAndRecentMatches() # get both Upcoming & Recent matches (faster than running them separately)
    # these matches are displayed on cev.eu under "GameHub"
    print( matches )

    x = await matches[0].toMatch() # any CalendarMatch object can be cast to a "full" Match (full capabilities), a match centre link is currently required though
    # if no match centre link exists for this match, toMatch returns None
    if x:
        print( await x.cache() )

asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
asyncio.run(main())



# -*- coding: utf-8 -*-
"""cevlib"""
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from abc import ABC, abstractmethod
from typing import Any, Coroutine, Dict, List, Union


JObject = Dict[str, Any]
JArray = List[Dict[str, Any]]

class IType(ABC):
    """model interface"""
    @property
    @abstractmethod
    def valid(self) -> bool:
        """is valid?"""

    def __bool__(self) -> bool:
        return self.valid

    @abstractmethod
    def toJson(self) -> Union[JObject,
                          JArray,
                          List[JArray],
                          Coroutine[Any, Any, JObject],
                          Coroutine[Any, Any, JArray]]:
        """serialise"""



import asyncio

from cevlib.match import Match

import json

async def main() -> None:
    match = await Match.byUrl("https://www.cev.eu/match-centres/2022-european-cups/cev-volleyball-challenge-cup-2022-women/chcw-78-rc-cannes-v-savino-del-bene-scandicci/")

    x = await match.toJson()
    with open("sample.json", "w") as outfile:
        jdata = json.dumps(x, indent=4)
        outfile.write(jdata)

asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
asyncio.run(main())



import os

import asyncio

from cevlib.match import Match, MatchCache

async def main():
    match = await Match.byUrl("https://championsleague.cev.eu/en/match-centres/cev-champions-league-volley-2022/women/clw-38-lp-salo-v-asptt-mulhouse-vb/")

    cache = await match.cache()
    # OR
    cache = await MatchCache.fromMatch(match)

    print("playByPlay", cache.playByPlay)
    print("competition", cache.competition)
    print("topPlayers", cache.topPlayers)
    print("report", cache.report)
    print("info", cache.info)
    print("gallery", cache.gallery) # Disclaimer: Photos featured on the CEV Photo Galleries are downloadable copyright free for media purposes only and only if CEV is credited as the source material. They are protected by copyright for all other commercial purposes. Those wishing to use CEV Photo Gallery photos for other commercial purposes should contact press@cev.eu
    print("matchCentreLink", cache.matchCentreLink)
    print("currentScore", cache.result)
    print("duration", cache.duration)
    print("startTime", cache.startTime)
    print("venue", cache.venue)
    print("homeTeam", cache.homeTeam)
    print("watchLink", cache.watchLink)
    print("highlightsLink", cache.highlightsLink)
    print("state", cache.state)

if os.name == "nt": # windows only
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
asyncio.run(main())



import asyncio

from cevlib.match import Match

from time import time

async def main() -> None:
    # cache up to 300ms

    match = await Match.byUrl("https://championsleague.cev.eu/en/match-centres/cev-champions-league-volley-2022/men/clm-61-cucine-lube-civitanova-v-ok-merkur-maribor/")
    match2 = await Match.byUrl("https://championsleague.cev.eu/en/match-centres/cev-champions-league-volley-2022/men/clm-61-cucine-lube-civitanova-v-ok-merkur-maribor/")

    t1 = time()
    cache = await match.cache()
    print(cache.venue)
    print(cache.result)
    print(cache.homeTeam)
    print(cache.competition)
    t2 = time()

    print(await match2.init())
    print(await match2.venue())
    print(await match2.result())
    print(await match2.homeTeam())
    print(await match2.competition())
    t3 = time()

    print(t2 - t1)
    print(t3 - t2)

    # BUT
    # cache up to 1s faster

    match = await Match.byUrl("https://championsleague.cev.eu/en/match-centres/cev-champions-league-volley-2022/men/clm-61-cucine-lube-civitanova-v-ok-merkur-maribor/")
    match2 = await Match.byUrl("https://championsleague.cev.eu/en/match-centres/cev-champions-league-volley-2022/men/clm-61-cucine-lube-civitanova-v-ok-merkur-maribor/")

    t1 = time()
    cache = await match.cache()
    print(cache.result)
    print(cache.homeTeam)
    print(cache.report)
    t2 = time()

    print(await match2.init())
    print(await match2.result())
    print(await match2.homeTeam())
    print(await match2.report())
    t3 = time()

    print(t2 - t1)
    print(t3 - t2)

    return

asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
asyncio.run(main())



# -*- coding: utf-8 -*-
"""cevlib"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from typing import List, Optional

from cevlib.helpers.dictTool import DictEx

from cevlib.types.iType import IType, JArray, JObject
from cevlib.types.types import Position, TeamStatisticType, TopPlayerType


class TeamStatistic(IType):
    """a team's single stat (from one set)"""
    def __init__(self, data: JObject, home: bool) -> None:
        dex = DictEx(data)
        self._type = TeamStatisticType.parse(dex.ensure("Name", str))
        self._value = dex.ensure("HomeTeamValue" if home else "AwayTeamValue", int)
        self._percent = dex.ensure("HomeTeamPercent" if home else "AwayTeamPercent", float)

    def toJson(self) -> JObject:
        return {
            "type": self.type.value,
            "value": self._value,
            "percent": self._percent
        }

    @property
    def type(self) -> TeamStatisticType:
        """type of the statistic"""
        return self._type

    @property
    def value(self) -> int:
        """absolute value"""
        return self._value

    @property
    def percent(self) -> float:
        """relative value (0 - 100)"""
        return self._percent

    @property
    def valid(self) -> bool:
        return None not in (self._type, self._value)

    def __repr__(self) -> str:
        return f"(cevlib.types.stats.TeamStatistic) {self._type} {self._value} ({self._percent}%)"


class TeamStatisticSet(IType):
    """a team's stats of one set"""
    def __init__(self, data: JArray, name: str, home: bool) -> None:
        self._stats = [ TeamStatistic(statistic, home) for statistic in data ]
        self._name = name

    def toJson(self) -> JObject:
        return {
            "stats": [ stat.toJson() for stat in self.stats ],
            "name": self.name
        }

    @property
    def valid(self) -> bool:
        return self._name is not None

    @property
    def stats(self) -> List[TeamStatistic]:
        """all stats"""
        return self._stats

    @property
    def name(self) -> str:
        """name of the type"""
        return self._name

    def byType(self, type_: TeamStatisticType) -> Optional[TeamStatistic]:
        """determine by type"""
        for stat in self._stats:
            if stat.type == type_:
                return stat
        return None

    def __repr__(self) -> str:
        return f"(cevlib.types.stats.TeamStatisticSet) {self._name} {self._stats}"


class TeamStatistics(IType):
    """a team's stats"""
    def __init__(self, data: JObject, home: bool) -> None:
        tabs = data.get("Tabs") or [ ]
        self._setStats: List[TeamStatisticSet] = [ ]
        for tab in tabs:
            self._setStats.append(TeamStatisticSet(tab.get("Statistics"), tab.get("Name"), home))

    @property
    def setStats(self) -> List[TeamStatisticSet]:
        """stats per set"""
        return self._setStats

    @property
    def valid(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"(cevlib.types.stats.TeamStatistics) {self._setStats}"

    def toJson(self) -> JObject:
        return {
            "setStats": [ setStat.toJson() for setStat in self._setStats ]
        }


class PlayerStatistic(IType):
    """a player's stats"""
    def __init__(self, data: JObject) -> None:
        dex = DictEx(data)
        self._points: int = dex.ensure("Points", int)
        self._serves: int = dex.ensure("Serves", int)
        self._spikes: int = dex.ensure("Spikes", int)
        self._blocks: int = dex.ensure("Blocks", int)
        self._receptions: int = dex.ensure("Reception", int)
        self._spikePerc: int = int(dex.ensure("SpikePerc", str).replace("%", ""))
        self._receptionPerc: int = int(dex.ensure("PositiveReceptionPerc", str).replace("%", ""))

    @property
    def valid(self) -> bool:
        return None not in (self._points,
                            self._serves,
                            self._spikes,
                            self._blocks,
                            self._receptions,
                            self._spikePerc,
                            self._receptionPerc)

    def toJson(self) -> JObject:
        return {
            "points": self.points,
            "serves": self.serves,
            "spikes": self.spikes,
            "blocks": self.blocks,
            "receptions": self.receptions,
            "spikePercentage": self.spikePercentage,
            "receptionPercentage": self.receptionPercentage
        }

    @property
    def points(self) -> int:
        """points scored"""
        assert self.valid
        return self._points

    @property
    def serves(self) -> int:
        """serves"""
        assert self.valid
        return self._serves

    @property
    def spikes(self) -> int:
        """spikes"""
        assert self.valid
        return self._spikes

    @property
    def blocks(self) -> int:
        """blocks"""
        assert self.valid
        return self._blocks

    @property
    def receptions(self) -> int:
        """receptions"""
        assert self.valid
        return self._receptions

    @property
    def spikePercentage(self) -> int:
        """0 - 100"""
        assert self.valid
        return self._spikePerc

    @property
    def receptionPercentage(self) -> int:
        """0 - 100"""
        assert self.valid
        return self._receptionPerc

    def __repr__(self) -> str:
        return f"(cevlib.types.stats.PlayerStatistic) Spikes: {self._spikes} Blocks: {self._blocks} Points: {self._points} Serves: {self._serves} Receptions: {self._receptions}" # pylint: disable=line-too-long


class TopPlayerPlayer(IType):
    """one of the top players of a TopPlayerType"""
    def __init__(self, data: JObject) -> None:
        dex = DictEx(data)
        self._number: int = dex.ensure("Number", int)
        self._name: str = dex.ensure("Name", str)
        self._position = Position.parse(dex.ensure("Position", str))
        self._score: int = dex.ensure("Score", int)
        self._nationality: str = dex.ensure("Team", str)
        self._image: str = dex.ensure("Image", str)

    def toJson(self) -> JObject:
        return {
            "number": self.number,
            "name": self.name,
            "position": self.position.value,
            "score": self.score,
            "nationality": self.nationality,
            "image": self.image
        }

    @property
    def valid(self) -> bool:
        return None not in (self._number, self._name, self._position, self._score)

    @property
    def number(self) -> int:
        """player's number"""
        return self._number

    @property
    def score(self) -> int:
        """player's score"""
        return self._score

    @property
    def position(self) -> Position:
        """player's position"""
        return self._position

    @property
    def name(self) -> str:
        """player's name"""
        return self._name

    @property
    def nationality(self) -> str:
        """player's nationality"""
        return self._nationality

    @property
    def image(self) -> str:
        """player's image (link)"""
        return self._image

    def __repr__(self) -> str:
        return f"(cevlib.types.stats.TopPlayerPlayer) {self._name} ({self._number}) Score: {self._score} as {self._position}" # pylint: disable=line-too-long


class TopPlayer(IType):
    """top players of a TopPlayerType"""
    def __init__(self, data: JObject) -> None:
        dex = DictEx(data)
        self._type = TopPlayerType.parse(dex.ensure("Type", str))
        self._players: List[TopPlayerPlayer] = [ ]
        for player in dex.ensure("Match", DictEx).ensure("Players", list):
            self._players.append(TopPlayerPlayer(player))

    def toJson(self) -> JObject:
        return {
            "type": self._type.value,
            "players": [ player.toJson() for player in self._players ]
        }

    @property
    def valid(self) -> bool:
        return bool(self._players)

    @property
    def type(self) -> TopPlayerType:
        """type"""
        return self._type

    @property
    def players(self) -> List[TopPlayerPlayer]:
        """top players of this type"""
        return self._players

    @property
    def topPlayer(self) -> TopPlayerPlayer:
        """best (top) player of this type"""
        return self._players[0]

    def __repr__(self) -> str:
        return f"(cevlib.types.stats.TopPlayer) {self._type} {self._players}"


class TopPlayers(IType):
    """top players"""
    def __init__(self) -> None:
        self._topPlayers: List[TopPlayer] = [ ]

    def toJson(self) -> JObject:
        return {
            "topPlayers": [ player.toJson() for player in self._topPlayers ]
        }

    def topPlayers(self) -> List[TopPlayer]:
        """all top players of all types"""
        return self._topPlayers

    def append(self, topPlayer: TopPlayer) -> None:
        """appends a top player type"""
        for player in self._topPlayers:
            if player.type == topPlayer.type:
                return
        self._topPlayers.append(topPlayer)

    @property
    def valid(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"(cevlib.types.stats.TopPlayers) {self._topPlayers}"



# -*- coding: utf-8 -*-
"""cevlib"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from enum import Enum

class TeamStatisticType(Enum):
    """all possible team statistics"""
    WinningSpikes = "winningSpikes"
    KillBlocks = "killBlocks"
    Aces = "aces"
    OpponentErrors = "opponentErrors"
    Points = "points"
    Unknown = "unknown"

    @staticmethod
    def parse(value: str) -> TeamStatisticType:
        """parse by value"""
        if value == "Winning Spikes":
            return TeamStatisticType.WinningSpikes
        if value == "Kill Blocks":
            return TeamStatisticType.KillBlocks
        if value == "Aces":
            return TeamStatisticType.Aces
        if value == "Opponent Errors":
            return TeamStatisticType.OpponentErrors
        if value == "Points":
            return TeamStatisticType.Points
        return TeamStatisticType.Unknown


class TopPlayerType(Enum):
    """all 'top player' awards"""
    Scorer = "scorer"
    Attacker = "attacker"
    Blocker = "blocker"
    Server = "server"
    Receiver = "receiver"
    Unknown = "unknown"

    @staticmethod
    def parse(value: str) -> TopPlayerType:
        """parse by value"""
        if value == "Scorer":
            return TopPlayerType.Scorer
        if value == "Attacker":
            return TopPlayerType.Attacker
        if value == "Blocker":
            return TopPlayerType.Blocker
        if value == "Server":
            return TopPlayerType.Server
        if value == "Receiver":
            return TopPlayerType.Receiver
        return TopPlayerType.Unknown


class CompetitionGender(Enum):
    """gender of the competition"""
    Women = "Women"
    Men = "Men"
    Unknown = "Unknown"

    @staticmethod
    def parse(value: str) -> CompetitionGender:
        """parse by value"""
        if value in ("Women", "W"):
            return CompetitionGender.Women
        if value in ("Men", "M"):
            return CompetitionGender.Men
        return CompetitionGender.Unknown


class Position(Enum):
    """all player positions"""
    Setter = "setter"
    MiddleBlocker = "middleBlocker"
    OutsideSpiker = "outsideSpiker"
    Opposite = "opposite"
    Libero = "libero"
    HeadCoach = "headCoach"
    Unknown = "unknown"

    @staticmethod
    def parse(value: str) -> Position:
        """parse by value"""
        if value == "Setter":
            return Position.Setter
        if value == "Middle blocker":
            return Position.MiddleBlocker
        if value == "Outside spiker":
            return Position.OutsideSpiker
        if value == "Opposite":
            return Position.Opposite
        if value == "Libero":
            return Position.Libero
        if value == "Head Coach":
            return Position.HeadCoach
        return Position.Unknown


class Zone(Enum):
    """all zones"""
    One = 1
    Two = 2
    Three = 3
    Four = 4
    Five = 5
    Six = 6
    Sub = 0
    Featured = 7
    Unknown = -1

    @staticmethod
    def parse(value: int) -> Zone:
        """parse by value"""
        if value == 1:
            return Zone.One
        if value == 2:
            return Zone.Two
        if value == 3:
            return Zone.Three
        if value == 4:
            return Zone.Four
        if value == 5:
            return Zone.Five
        if value == 6:
            return Zone.Six
        if value in (7, 8):
            return Zone.Featured
        if value == 0:
            return Zone.Sub
        return Zone.Unknown


class PlayType(Enum):
    """all types of 'play by play'"""
    Spike = "spike"
    Serve = "serve"
    Block = "block"
    FirstServe = "firstServe"
    Unknown = "unknown"

    @staticmethod
    def parse(typeName: str) -> PlayType:
        """parse by value"""
        if typeName == "Spike":
            return PlayType.Spike
        if typeName == "Serve":
            return PlayType.Serve
        if typeName == "Block":
            return PlayType.Block
        if typeName == "First Serve":
            return PlayType.FirstServe
        return PlayType.Unknown


class MatchState(Enum):
    """state of the match"""
    Upcoming = "upcoming"
    Live = "live"
    Finished = "finished"
    Unknown = "unknown"

    @staticmethod
    def parse(started: bool, finished: bool) -> MatchState:
        """determine"""
        if not started and not finished:
            return MatchState.Upcoming
        if started and not finished:
            return MatchState.Live
        if finished:
            return MatchState.Finished
        return MatchState.Unknown



# -*- coding: utf-8 -*-
"""cevlib"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

# (IMatch can be property or async)
# pylint: disable=invalid-overridden-method

from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

import aiohttp

from cevlib.match import Match
from cevlib.helpers.dictTool import DictEx
from cevlib.types.competition import MatchCompetition
from cevlib.types.iMatch import IMatch
from cevlib.types.iType import IType, JObject
from cevlib.types.results import Result
from cevlib.types.team import Team
from cevlib.types.types import MatchState


class CalendarMatch(IMatch):
    """simplified match"""
    def __init__(self,
                 url: Optional[str],
                 competition: MatchCompetition,
                 homeTeam: Team,
                 awayTeam: Team,
                 venue: str,
                 startTime: str,
                 result: Result,
                 finished: bool) -> None:
        self._matchCentreLink = url
        self._competition = competition
        self._homeTeam = homeTeam
        self._awayTeam = awayTeam
        self._venue = venue
        if not startTime.endswith("Z"):
            startTime += "Z"
        self._startTime = datetime.strptime(startTime, "%Y-%m-%dT%H:%M:%SZ")
        duration = (datetime.now() - self._startTime)
        self._finished = True if (self._startTime.year == 1900 or duration.days) else finished
        self._result = result
        self._state = MatchState.parse(datetime.utcnow() >= self._startTime, self._finished)

    @staticmethod
    def parse(data: JObject) -> CalendarMatch:
        """parses from json object"""
        dex = DictEx(data)
        return CalendarMatch(dex.ensureString("MatchCentreUrl"),
                             MatchCompetition({ "Competition": dex.ensureString("CompetitionName"),
                                           "CompetitionLogo": dex.ensureString("CompetitionLogo"),
                                           "Phase": dex.ensureString("PhaseName") }),
                             Team.build( dex.ensureString("HomeTeamName"),
                                         dex.ensureString("HomeTeamLogo"),
                                         dex.ensureString("HomeClubCode"),
                                         True ),
                             Team.build( dex.ensureString("GuestTeamName"),
                                         dex.ensureString("GuestTeamLogo"),
                                         dex.ensureString("GuestClubCode"),
                                         False ),
                             dex.ensureString("StadiumName"),
                             dex.ensureString("MatchDateTime_UTC"),
                             Result({
                                "homeSetsWon": dex.ensureInt("WonSetHome"),
                                "awaySetsWon": dex.ensureInt("WonSetGuest"),
                             }),
                             dex.ensureBool("Finalized"))

    @staticmethod
    def shortcutMatch(competition: MatchCompetition, team: Team) -> CalendarMatch:
        """a match that has only one team. Used in case of direct qualification"""
        return CalendarMatch(None,
                             competition,
                             team,
                             Team.build("", "", "", False),
                             "",
                             datetime.fromtimestamp(0).strftime("%Y-%m-%dT%H:%M:%SZ"),
                             Result({}),
                             True)

    @property
    def teams(self) -> Tuple[Team, Team]:
        """(home, away)"""
        return ( self._homeTeam, self._awayTeam )

    async def toMatch(self) -> Optional[Match]:
        """casts this calendar match to a full match"""
        if not self._matchCentreLink:
            return None
        return await Match.byUrl(self._matchCentreLink)

    def __repr__(self) -> str:
        return f"(cevlib.calendar.CalendarMatch) {self._matchCentreLink} {self._competition} {self._venue} {self._startTime}\n{self._homeTeam}\n{self._awayTeam}\n{self._result}" # pylint: disable=line-too-long

    def toJson(self) -> Dict[str, Any]:
        return {
            "competition": self.competition.toJson(),
            "homeTeam": self.homeTeam.toJson(),
            "awayTeam": self.awayTeam.toJson(),
            "venue": self.venue,
            "startTime": str(self.startTime),
            "result": self.result.toJson(),
            "matchCentreLink": self.matchCentreLink,
            "state": self._state.value
        }

    @property
    def state(self) -> MatchState:
        return self._state

    @property
    def competition(self) -> MatchCompetition:
        return self._competition

    @property
    def homeTeam(self) -> Team:
        return self._homeTeam

    @property
    def awayTeam(self) -> Team:
        return self._awayTeam

    @property
    def venue(self) -> str:
        return self._venue

    @property
    def matchCentreLink(self) -> Optional[str]:
        return self._matchCentreLink

    @property
    def startTime(self) -> datetime:
        return self._startTime

    @property
    def result(self) -> Result:
        return self._result

    @property
    def valid(self) -> bool:
        return True

    @property
    def finished(self) -> bool:
        return self._finished


class Calendar(IType):
    """calendar reader (static methods only)"""
    @staticmethod
    async def matchesOfMonth(month: Optional[int] = None,
                             year: Optional[int] = None) -> List[CalendarMatch]:
        """gets all matches of a month"""
        today = datetime.now()
        timestamp = datetime(year if year else today.year,
                             month if month else today.month, 1).strftime("%Y-%m-%dT%H:%M:%SZ")
        matches: List[Dict[str, Any]] = [ ]
        async with aiohttp.ClientSession() as client:
            async with client.get(f"https://www.cev.eu/umbraco/api/CalendarApi/GetCalendar?nodeId=11346&culture=en-US&date={timestamp}") as resp: # pylint: disable=line-too-long
                jdata = await resp.json(content_type=None)
                calendar = DictEx(jdata)
                for date in calendar.ensureList("Dates"):
                    matches.extend(date.get("Matches") or [ ])
        return [ CalendarMatch.parse(match)
                 for match in matches ]


    @staticmethod
    async def recentMatches() -> List[CalendarMatch]:
        """gets all matches that cev.eu displays as 'recent'"""
        matches = await Calendar._getLiveScoreMatches()
        return [ Calendar._liveScoresToCalendarMatch(match)
                 for match in matches
                 if match.get("matchState_String") == "FINISHED" ]

    @staticmethod
    async def upcomingMatches() -> List[CalendarMatch]:
        """gets all matches that cev.eu displays as 'upcoming'"""
        matches = await Calendar._getLiveScoreMatches()
        return [ Calendar._liveScoresToCalendarMatch(match)
                 for match in matches
                 if not match.get("matchState_String") == "FINISHED" ]

    @staticmethod
    async def upcomingAndRecentMatches() -> List[CalendarMatch]:
        """gets all matches that cev.eu displays as 'recent' or 'upcoming'"""
        matches = await Calendar._getLiveScoreMatches()
        return [ Calendar._liveScoresToCalendarMatch(match)
                 for match in matches ]

    @staticmethod
    def _liveScoresToCalendarMatch(match: Dict[str, Any]) -> CalendarMatch:
        dex = DictEx(match)
        compDict = dex.ensureDict("competition")
        compDict["Phase"] = dex.get("phaseName")
        compDict["Leg"] = dex.get("legName")
        compDict["GroupPool"] = dex.get("groupName")
        compDict["MatchNumber"] = dex.get("matchNumber")
        return CalendarMatch(dex.get("matchCentreLink"),
                             MatchCompetition(compDict),
                             Team.build( dex.ensureString("homeTeam"),
                                         dex.ensureString("homeTeamIcon"),
                                         dex.ensureString("homeTeamNickname"),
                                         True ),
                             Team.build( dex.ensureString("awayTeam"),
                                         dex.ensureString("awayTeamIcon"),
                                         dex.ensureString("awayTeamNickname"),
                                         False ),
                             dex.ensureString("matchLocation"),
                             dex.ensureString("utcStartDate"),
                             Result(match),
                             dex.ensureString("matchState_String") == "FINISHED")

    @staticmethod
    async def _getLiveScoreMatches() -> List[Dict[str, Any]]:
        matches = [ ]
        async with aiohttp.ClientSession() as client:
            async with client.get("https://www.cev.eu/LiveScores.json") as resp:
                jdata = DictEx(await resp.json(content_type=None))
                for competition in jdata.ensure("competitions", list):
                    dex = DictEx(competition)
                    for match in dex.ensure("matches", list):
                        match["competition"] = { "Competition": dex.tryGet("competitionName",
                                                                              str),
                                                 "id": dex.tryGet("competitionId", str) }
                        matches.append(match)

        return matches

    def __repr__(self) -> str:
        return "(cevlib.calendar.Calendar)"

    def toJson(self) -> Dict[str, Any]:
        return { }

    @property
    def valid(self) -> bool:
        return True



# -*- coding: utf-8 -*-
"""cevlib"""
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from typing import List, Optional
import re

from bs4 import BeautifulSoup # type: ignore
from bs4.element import Tag # type: ignore

from cevlib.types.iType import IType, JObject


class Referee(IType):
    """referee"""
    def __init__(self, tag: Tag) -> None:
        self._type = tag.find("div", class_="u-border-grey-light").string
        imgStyle = tag.find("div", class_="accordion-content__image").get("style")
        imgMatch: Optional[re.Match[str]] = re.search(r"https:\/\/([\w_-]+(?:(?:\.[\w_-]+)+))([\w.,@;?^=%&:\/~+#-]*Images\/Officials\/[\w .,@;?^=%&:\/~+#-]*[\w@?^=%&\/~+#-])", # pylint: disable=line-too-long
                                                      imgStyle)
        self._img = imgMatch.group(0) if imgMatch else None

        nameAndNat = tag.find("div", class_="accordion-content__item")
        name = nameAndNat.find("div")

        self._name = name.string
        name.decompose()

        nat = nameAndNat.find("div")
        self._nationality = nat.string

    def __repr__(self) -> str:
        return f"(cevlib.types.info.Referee) {self._name} {self._type} ({self._nationality}) {self._img}" # pylint: disable=line-too-long

    @property
    def valid(self) -> bool:
        return True

    def toJson(self) -> JObject:
        return {
            "name": self._name,
            "image": self._img,
            "type": self._type,
            "nationality": self._nationality
        }


class Venue(IType):
    """venue info"""
    def __init__(self, tag: Tag) -> None:
        self._img = tag.find("img", class_="u-object-cover").get("src")
        data = tag.find("p")
        cap = data.find("span")
        cap.strong.decompose()
        self._capacity = cap.get_text(strip = True)
        cap.decompose()
        self._name = data.get_text(strip = True)

    def __repr__(self) -> str:
        return f"(cevlib.types.info.Venue) {self._name} (capacity: {self._capacity}) {self._img}"

    @property
    def valid(self) -> bool:
        return True

    def toJson(self) -> JObject:
        return {
            "name": self._name,
            "image": self._img,
            "capacity": self._capacity
        }


def _getAccordionTitle(tag: Tag) -> Optional[str]:
    anchor = tag.find("a", class_="accordion-title")
    if not anchor:
        return None
    title = anchor.string
    if isinstance(title, str):
        return title
    return None


class Info(IType):
    """match info"""
    def __init__(self, html: str) -> None:
        soup = BeautifulSoup(html, "html.parser")
        infoText = soup.find("div", class_="text-container")
        self._infoText = infoText.get_text(strip=True, separator='<br>') if infoText else None
        self._officials: List[Referee] = [ ]
        self._venue: Optional[Venue] = None

        infoContainer = soup.find("div", class_="match-info")
        if not infoContainer:
            return
        infoItems = infoContainer.find_all("li", class_="accordion-item")
        for infoItem in infoItems:
            if _getAccordionTitle(infoItem) == "Officials":
                for official in infoItem.find_all("div", class_="u-flex-1"):
                    self._officials.append(Referee(official))
            if _getAccordionTitle(infoItem) == "How To Attend":
                self._venue = Venue(infoItem)

    def __repr__(self) -> str:
        return f"(cevlib.types.info.Info) {self._infoText} {self._officials} {self._venue}"

    @property
    def valid(self) -> bool:
        return True

    def toJson(self) -> JObject:
        return {
            "infoText": self._infoText,
            "officials": [ official.toJson() for official in self._officials ],
            "venue": self._venue.toJson() if self._venue else None
        }



# -*- coding: utf-8 -*-
"""cevlib"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from typing import List, Optional, Union
from bs4 import BeautifulSoup # type: ignore
from bs4.element import Tag, NavigableString # type: ignore

from cevlib.types.iType import IType, JObject


class MatchReport(IType):
    """match report"""
    def __init__(self, html: str) -> None:
        soup = BeautifulSoup(html, "html.parser")
        self._quotes: List[MatchQuote] = [ ]
        self._inNumbers: List[MatchInNumber]  = [ ]
        self._body: Optional[str] = None
        self._headline: Optional[str] = None
        try:
            matchReport = soup.find("div", class_="match-report")
            try:
                self._headline = str(matchReport.h2.get_text(strip=True))
                self._body = str(matchReport.find("div",
                                                  class_="match-report__summary-container").p.text)
            except Exception:
                self._headline = None
                self._body = None

            matchReport.decompose()
            matchReport = soup.find("div", class_="match-report")
            self._quotes = MatchQuote.parse(soup)
            self._inNumbers = MatchInNumber.parse(matchReport)
        except AttributeError:
            self._headline = None
            self._body = None
            self._quotes = [ ]
            self._inNumbers = [ ]

    def toJson(self) -> JObject:
        return {
            "headline": self.headline,
            "body": self.body,
            "quotes": [ quote.toJson() for quote in self.quotes ],
            "inNumbers": [ inNumber.toJson() for inNumber in self.inNumbers ]
        }

    @property
    def headline(self) -> Optional[str]:
        """headline"""
        assert self.valid
        return self._headline

    @property
    def body(self) -> Optional[str]:
        """body content"""
        assert self.valid
        return self._body

    @property
    def quotes(self) -> List[MatchQuote]:
        """match quotes"""
        assert self.valid
        return self._quotes

    @property
    def inNumbers(self) -> List[MatchInNumber]:
        """fancy stats (e.g. minutes played, attendance, ...)"""
        assert self.valid
        return self._inNumbers

    @property
    def valid(self) -> bool:
        return None not in (self._headline, self._body) and\
               len(self._quotes) > 0 or \
               len(self._inNumbers) > 0

    def __repr__(self) -> str:
        return f"(cevlib.types.report.MatchQuote) {self._headline} ({self._body})\n{self._quotes}\n{self._inNumbers}" # pylint: disable=line-too-long


class MatchInNumber(IType):
    """fancy stats (e.g. minutes played, attendance, ...)"""
    def __init__(self, value: str, title: str, description: str) -> None:
        self._value = value
        self._title = title
        self._description = description

    def toJson(self) -> JObject:
        return {
            "title": self.title,
            "description": self.description,
            "value": self.value
        }

    @property
    def valid(self) -> bool:
        return None not in (self._value, self._title, self._description)

    @property
    def value(self) -> str:
        """the 'number'"""
        assert self.valid
        return self._value

    @property
    def title(self) -> str:
        """stat name"""
        assert self.valid
        return self._title

    @property
    def description(self) -> str:
        """description/explanation"""
        assert self.valid
        return self._description

    def __repr__(self) -> str:
        return f"(cevlib.types.report.MatchInNumber) {self._title}: {self._value} ({self._description})" # pylint: disable=line-too-long

    @staticmethod
    def parse(soup: Union[Tag, NavigableString]) -> List[MatchInNumber]:
        """parse from html tag"""
        numbers = soup.find_all("div", class_="column-container")
        matchInNumbers: List[MatchInNumber] = [ ]
        for number in numbers:
            title = number.find("span", class_="col__content-title").get_text(strip=True)
            matchInNumbers.append(MatchInNumber(number.span.text,
                                                title,
                                                number.p.text))
        return matchInNumbers


class MatchQuote(IType):
    """match quote"""
    def __init__(self, quote: str, cite: str) -> None:
        self._quote = quote.replace("“", "").replace("”", "").removeprefix(" ").removesuffix(" ")
        self._cite = cite.split("<br>")[0]
        self._citeDescription = cite.split("<br>")[1]

    def toJson(self) -> JObject:
        return {
            "quote": self.quote,
            "cite": self.cite,
            "citeDescription": self.citeDescription
        }

    @property
    def quote(self) -> str:
        """the quote itself"""
        assert self.valid
        return self._quote

    @property
    def cite(self) -> str: # TODO rename to source?
        """cite/source"""
        assert self.valid
        return self._cite

    @property
    def citeDescription(self) -> str:
        """description of the cite/source"""
        assert self.valid
        return self._citeDescription

    @property
    def valid(self) -> bool:
        return None not in (self._quote, self._cite, self._citeDescription)

    def __repr__(self) -> str:
        return f"(cevlib.types.report.MatchQuote) {self._cite} ({self._citeDescription}): '{self._quote}'" # pylint: disable=line-too-long

    @staticmethod
    def parse(soup: BeautifulSoup) -> List[MatchQuote]:
        """parses a quote from html"""
        quotes: List[MatchQuote] = [ ]
        quoteElements = soup.find_all(class_="quote-block")
        for element in quoteElements:
            quotes.append(MatchQuote(element.p.text,
                                     element.cite.get_text(strip=True, separator='<br>')))
        return quotes



# -*- coding: utf-8 -*-
"""cevlib"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

import re
from typing import List, Optional

from cevlib.helpers.dictTool import DictEx, ListEx

from cevlib.types.iType import IType, JArray, JObject


class SetResult(IType):
    """result of a single set"""
    def __init__(self, data: JObject) -> None:
        dex = DictEx(data)
        self._homeScore: int = dex.ensure("homeScore", int)
        self._awayScore: int = dex.ensure("awayScore", int)
        self._setNumber: int = dex.ensure("setNumber", int)
        self._isInPlay: bool = dex.ensure("isInPlay", bool)

    def toJson(self) -> JObject:
        return {
            "homeScore": self.homeScore,
            "awayScore": self.awayScore,
            "setNumber": self.setNumber,
            "isInPlay": self.isInPlay
        }

    @property
    def valid(self) -> bool:
        return None not in (self._homeScore, self._awayScore, self._setNumber, self._isInPlay)\
            and 0 not in (self._homeScore, self._awayScore)

    @property
    def homeScore(self) -> int:
        """home team score"""
        return self._homeScore

    @property
    def awayScore(self) -> int:
        """away team score"""
        return self._awayScore

    @property
    def setNumber(self) -> int:
        """set number (1 - 5)"""
        return self._setNumber

    @property
    def isInPlay(self) -> bool:
        """is in play?"""
        return self._isInPlay

    def __repr__(self) -> str:
        return f"(cevlib.types.results.setResult) {self._homeScore} - {self._awayScore} ({'ongoing' if self._isInPlay else 'finished'})" # pylint: disable=line-too-long

    @staticmethod
    def parseList(setResults: JArray) -> List[SetResult]:
        """parse result from set results"""
        results = [ ]
        for setResult in setResults:
            result = SetResult(setResult)
            if result:
                results.append(result)
        return results

    @staticmethod
    def parseFromPlayByPlay(data: JObject) -> SetResult:
        """parse result from play by play"""
        dex = DictEx(data)
        score = ListEx(dex.ensure("Description", str).split("-"))
        return SetResult({
            "homeScore": score.ensure(0, int),
            "awayScore": score.ensure(1, int),
            "setNumber": dex.ensure("SetNumber", int),
            "isInPlay": False
        })


class Result(IType):
    """full result"""
    def __init__(self, data: JObject) -> None:
        dex = DictEx(data)
        self._sets: List[SetResult] = SetResult.parseList(data.get("setResults") or [ ])
        currentSet = SetResult(dex.ensure("currentSetScore", dict))
        if currentSet:
            self._sets.append(currentSet)
        self._hasGoldenSet: bool = dex.ensure("hasGoldenSet", bool)
        self._homeScore: int = dex.ensure("homeSetsWon", int)
        self._awayScore: int = dex.ensure("awaySetsWon", int)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Result):
            return False
        return self.toJson() == other.toJson()

    def toJson(self) -> JObject:
        return {
            "hasGoldenSet": self.hasGoldenSet,
            "sets": [ set_.toJson() for set_ in self.sets ],
            "regularSets": [ set_.toJson() for set_ in self.regularSets ],
            "goldenSet": self.goldenSet.toJson() if self.goldenSet else None,
            "homeScore": self.homeScore,
            "awayScore": self.awayScore
        }

    @staticmethod
    def parseFromForm(data: JObject) -> Result:
        """parses from form match"""
        dex = DictEx(data)
        sets = re.sub(r"[(</span>) ]", "", dex.ensure("SetsFormatted", str)).split(",")
        return Result({
            "homeSetsWon": dex.ensure("HomeTeam", DictEx).ensure("Score", int),
            "awaySetsWon": dex.ensure("AwayTeam", DictEx).ensure("Score", int),
            "setResults": [ {
                "homeScore": ListEx(set.split("-")).ensure(0, int),
                "awayScore": ListEx(set.split("-")).ensure(1, int),
                "setNumber": i + 1,
                "isInPlay": False
            } for (i, set) in enumerate(sets) if not set == "" ]
        })

    @property
    def sets(self) -> List[SetResult]:
        """all sets"""
        return self._sets

    @property
    def regularSets(self) -> List[SetResult]:
        """all regular sets (no golden set)"""
        if self.hasGoldenSet:
            return self.sets[:-1]
        return self.sets

    @property
    def goldenSet(self) -> Optional[SetResult]:
        """golden set (if one exists)"""
        if not self.hasGoldenSet:
            return None
        return self.sets[-1]

    @property
    def hasGoldenSet(self) -> bool:
        """has golden set"""
        return self._hasGoldenSet

    @property
    def latestSet(self) -> Optional[SetResult]:
        """latest set (that might still be active)"""
        if len(self.sets) == 0:
            return None
        return self.sets[len(self.sets) - 1]

    @property
    def homeScore(self) -> int:
        """home team score"""
        return self._homeScore

    @property
    def awayScore(self) -> int:
        """away team score"""
        return self._awayScore

    @property
    def valid(self) -> bool:
        return None not in (self._sets, self._homeScore, self._awayScore)

    def __repr__(self) -> str:
        return f"(cevlib.types.results.Result) {self._homeScore}:{self._awayScore} (Golden Set: {self._hasGoldenSet}) {self._sets}" # pylint: disable=line-too-long



# -*- coding: utf-8 -*-
"""cevlib"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from datetime import datetime
from typing import List, Optional

from cevlib.helpers.dictTool import DictEx, ListEx

from cevlib.types.iType import IType, JArray, JObject
from cevlib.types.matchPoll import TeamPoll
from cevlib.types.stats import PlayerStatistic, TeamStatistics
from cevlib.types.types import Position, Zone
from cevlib.types.results import Result


class Player(IType):
    """player model"""
    def __init__(self, data: JObject, playerStatsData: JArray) -> None:
        dex = DictEx(data)
        self._number = dex.ensure("Number", int)
        self._name = dex.ensure("Name", str, "N/A").title()
        self._position = Position.parse(dex.ensure("Position", str))
        self._image = dex.ensure("Image", str)
        self._isCaptain = dex.ensure("isCaptain", bool)
        self._zone = Zone.parse(dex.ensure("PositionNumber", int))
        self._id = dex.ensure("PlayerId", int)
        self._stats: Optional[PlayerStatistic] = None
        for value in playerStatsData:
            player = DictEx(value)
            if self._name.split(" ")[0] in player.ensure("Name", str).title() and \
               player.ensure("PlayerNumber", int) == self._number:
                self._stats = PlayerStatistic(player)
                break

    @property
    def valid(self) -> bool:
        return self._id is not None

    def toJson(self) -> JObject:
        return {
            "zone": self.zone.value,
            "position": self.position.value,
            "name": self.name,
            "id": self.id,
            "number": self.number,
            "stats": self._stats.toJson() if self._stats else None
        }

    @property
    def zone(self) -> Zone:
        """the player's zone"""
        return self._zone

    @property
    def position(self) -> Position:
        """the player's position"""
        return self._position

    @property
    def name(self) -> str:
        """the player's name"""
        return self._name

    @property
    def id(self) -> int:
        """the player's id"""
        return self._id

    @property
    def number(self) -> int:
        """the player's number"""
        return self._number

    @property
    def stats(self) -> Optional[PlayerStatistic]:
        """the player's stats"""
        return self._stats

    def __repr__(self) -> str:
        return f"(cevlib.types.team.Player) {self._name} ({self._number}/{self._id}) {self._position} ({self._zone})" # pylint: disable=line-too-long


class FormMatch(IType):
    """a very minimal match that is part of a team's 'form'"""
    def __init__(self,
                 won: bool,
                 link: str,
                 homeTeam: Team,
                 awayTeam: Team,
                 result: Result,
                 startTime: datetime) -> None:
        self._won = won
        self._link = link
        self._homeTeam = homeTeam
        self._awayTeam = awayTeam
        self._result = result
        self._startTime = startTime

    def toJson(self) -> JObject:
        return {
            "won": self.won,
            "link": self.link,
            "homeTeam": self._homeTeam.toJson(),
            "awayTeam": self._awayTeam.toJson(),
            "result": self._result.toJson(),
            "startTime": str(self._startTime)
        }

    @property
    def won(self) -> bool:
        """game win"""
        return self._won

    @property
    def link(self) -> str:
        """match centre url"""
        return self._link

    @staticmethod
    def parse(data: JObject) -> List[FormMatch]:
        """parses a form match"""
        matches: List[FormMatch] = [ ]
        for (index, match) in enumerate(data.get("Matches") or [ ]):
            if index >= len(data["RecentForm"]):
                break
            matches.append(FormMatch(data["RecentForm"][index],
                                    match["MatchCentreUrl"],
                                    Team.build(match["HomeTeam"]["Name"],
                                                match["HomeTeam"]["Logo"]["Url"],
                                                "N/A", True),
                                    Team.build(match["AwayTeam"]["Name"],
                                                match["AwayTeam"]["Logo"]["Url"],
                                                "N/A", False),
                                    Result.parseFromForm(match),
                                    datetime.strptime(match["MatchDateTime"], "%Y-%m-%dT%H:%M:%S")))
        return matches

    @property
    def valid(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"(cevlib.types.team.FormMatch) {self._won} ({self._link})"


class Team(IType):
    """team"""
    def __init__(self,
            data: JObject,
            playerStatsData: JObject,
            stats: TeamStatistics,
            matchPollData: JArray,
            form: JObject,
            icon: Optional[str] = None,
            nickname: Optional[str] = None,
            id_: int = 0) -> None:
        dex = DictEx(data)
        pollData = ListEx(matchPollData)
        self._stats = stats
        playerStatsList: JArray = [ ]
        for team in playerStatsData.get("Teams") or [ ]:
            playerStatsList.extend(team.get("Players"))

        teamLogo = dex.ensure("TeamLogo", DictEx)
        self._form = FormMatch.parse(form)
        self._nickname: Optional[str] = nickname
        self._name: Optional[str] = teamLogo.tryGet("AltText", str)
        self._logo: Optional[str] = icon or teamLogo.tryGet("Url", str)
        self._id: Optional[int] = id_ or dex.ensure("TeamId", int)

        self._poll: Optional[TeamPoll] = None
        if len(pollData) == 2:
            team = pollData.ensure(0, DictEx)
            if team.ensure("Id", int) != self._id:
                team = pollData.ensure(1, DictEx)
            self._poll = TeamPoll(team)

        self._players: List[Player] = [ ]
        if "TopLeftPlayer" in data: # assume that the others are as well, might improve
            self._players.append(Player(dex.ensure("TopLeftPlayer", dict), playerStatsList))
            self._players.append(Player(dex.ensure("TopMidPlayer", dict), playerStatsList))
            self._players.append(Player(dex.ensure("TopRightPlayer", dict), playerStatsList))
            self._players.append(Player(dex.ensure("BottomLeftPlayer", dict), playerStatsList))
            self._players.append(Player(dex.ensure("BottomMidPlayer", dict), playerStatsList))
            self._players.append(Player(dex.ensure("BottomRightPlayer", dict), playerStatsList))
            self._players.append(Player(dex.ensure("HeadCoach", dict), playerStatsList))
        self._players.extend([ Player(player, playerStatsList)
                               for player in dex.ensure("FeaturedPlayers", list) ])
        self._players.extend([ Player(player, playerStatsList)
                               for player in dex.ensure("SubPlayers", list) ])

    @staticmethod # TODO (swap with ctor (make parse))
    def build(name: str, icon: str, nickname: str, home: bool, id_: int = 0) -> Team:
        """builds a team"""
        return Team({ "TeamLogo": {
                            "AltText": name,
                            "Url": icon
                        }
                    }, { }, TeamStatistics({ }, home), [ ], { },
                    icon, nickname, id_)

    def toJson(self) -> JObject:
        return {
            "name": self.name,
            "nickname": self.nickname,
            "logo": self.logo,
            "id": self.id,
            "stats": self.stats.toJson(),
            "poll": self.poll.toJson() if self.poll else None,
            "form": [ form.toJson() for form in self.form ],
            "players": [ player.toJson() for player in self.players ]
        }

    @property
    def valid(self) -> bool:
        return None not in (self._name, self._id, self._stats)

    @property
    def form(self) -> List[FormMatch]:
        """the team's form"""
        return self._form

    def __repr__(self) -> str:
        return f"(cevlib.types.team.Team) {self._name} ({self._nickname}/{self._id}) \nplayers={self._players}\nform={self._form}" # pylint: disable=line-too-long

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Team):
            return False
        if self.id and other.id:
            return self.id == other.id
        return self.name == other.name

    @property
    def name(self) -> Optional[str]:
        """the team's name"""
        return self._name

    @property
    def nickname(self) -> Optional[str]:
        """the team's nickname (3 chars)"""
        return self._nickname

    @property
    def logo(self) -> Optional[str]:
        """the team's logo (link)"""
        return self._logo

    @property
    def id(self) -> Optional[int]:
        """the team's id"""
        return self._id

    @property
    def stats(self) -> TeamStatistics:
        """the team's stats"""
        return self._stats

    @property
    def poll(self) -> Optional[TeamPoll]:
        """the team's poll data"""
        return self._poll

    @property
    def players(self) -> List[Player]:
        """the team's players"""
        return self._players

    def getFirstPlayer(self,
                  zone: Optional[Zone] = None,
                  position: Optional[Position] = None,
                  id_: Optional[int] = None,
                  number: Optional[int] = None) -> Optional[Player]:
        """the team's first player that matches the filters"""
        players = self.getPlayers(zone, position, id_, number)
        if len(players) > 0:
            return players[0]
        return None

    def getPlayers(self,
                  zone: Optional[Zone] = None,
                  position: Optional[Position] = None,
                  id_: Optional[int] = None,
                  number: Optional[int] = None) -> List[Player]:
        """the team's filtered players"""
        eligiblePlayers = self._players
        if zone:
            eligiblePlayers = [ player
                                for player in eligiblePlayers
                                if player.zone == zone ]
        if position:
            eligiblePlayers = [ player
                                for player in eligiblePlayers
                                if player.position == position ]
        if id_ is not None:
            eligiblePlayers = [ player
                                for player in eligiblePlayers
                                if player.id == id_ ]
        if number is not None:
            eligiblePlayers = [ player
                                for player in eligiblePlayers
                                if player.number == number ]
        return eligiblePlayers



# -*- coding: utf-8 -*-
"""cevlib"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from typing import Optional

from cevlib.helpers.dictTool import DictEx

from cevlib.types.iType import IType, JObject
from cevlib.types.types import CompetitionGender


class MatchCompetition(IType):
    """a match's competition info"""
    def __init__(self, data: JObject) -> None:
        dex = DictEx(data)
        competition = dex.ensure("Competition", str)
        if "|" not in competition:
            competition += "|"
        self._name = competition.split("|")[0].removesuffix(" ")
        self._gender = CompetitionGender.parse(competition.split("|")[1]
                                                    .removeprefix(" ")
                                                    .split(" ", maxsplit = 1)[0])
        self._groupPool = dex.ensure("GroupPool", str)
        self._leg = dex.ensure("Leg", str)
        self._phase = dex.ensure("Phase", str)
        self._season = dex.ensure("Season", str)
        self._matchNumber = dex.ensure("MatchNumber", str)
        self._logo = dex.ensure("CompetitionLogo", str)

    @staticmethod
    # TODO might swap with ctor and replace with parse
    def buildPlain(name: Optional[str] = None,
                   gender: CompetitionGender = CompetitionGender.Unknown,
                   groupPool: Optional[str] = None,
                   leg: Optional[str] = None,
                   phase: Optional[str] = None,
                   season: Optional[str] = None,
                   matchNumber: Optional[str] = None,
                   logo: Optional[str] = None) -> MatchCompetition:
        """builds a competition"""
        return MatchCompetition({
            "Competition": f"{name} | {gender.value}",
            "GroupPool": groupPool,
            "Leg": leg,
            "Phase": phase,
            "Season": season,
            "MatchNumber": matchNumber,
            "CompetitionLogo": logo
        })

    @property
    def name(self) -> str:
        """competition name"""
        return self._name

    @property
    def displayName(self) -> str:
        """competition name + gender"""
        return f"{self._name} | {self._gender.value}"

    @property
    def gender(self) -> CompetitionGender:
        """competition gender"""
        return self._gender

    @property
    def groupPool(self) -> str:
        """competition group pool"""
        return self._groupPool

    @property
    def leg(self) -> str:
        """competition leg"""
        return self._leg

    @property
    def phase(self) -> str:
        """competition phase"""
        return self._phase

    @property
    def season(self) -> str:
        """season"""
        return self._season

    @property
    def matchNumber(self) -> str:
        """match number"""
        return self._matchNumber

    @property
    def logo(self) -> str:
        """competition logo (link)"""
        return self._logo

    def __repr__(self) -> str:
        return f"(cevlib.types.competition.Competition) {self.displayName} ({self._matchNumber}) {self._season} > {self._phase} > {self._groupPool} ({self._leg})" # pylint: disable=line-too-long

    @property
    def valid(self) -> bool:
        return None not in (self._name, self._gender)

    def toJson(self) -> JObject:
        return {
            "name": self.name,
            "displayName": self.displayName,
            "gender": self.gender.value,
            "groupPool": self.groupPool,
            "leg": self.leg,
            "phase": self.phase,
            "season": self.season,
            "logo": self.logo,
            "matchNumber": self.matchNumber
        }



import asyncio

from cevlib.match import Match
from cevlib.types.results import Result

async def main() -> None:
    match = await Match.byUrl("https://www.cev.eu/match-centres/2022-european-cups/cev-volleyball-cup-2022-women/ccw-54-mladost-zagreb-v-lks-commercecon-lodz/")

    await match.init()
    async def implement(x: Match, score: Result) -> None:
        print(score)
        print(await x.duration())

    match.addScoreObserver(implement)
    match.setScoreObserverInterval(10) # 10s

    while True:
        await asyncio.sleep(1)

asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
asyncio.run(main())



# -*- coding: utf-8 -*-
"""cevlib"""
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from abc import abstractmethod
from datetime import datetime, timedelta
from typing import Any, Coroutine, List, Optional, TypeVar, Union

from cevlib.types.competition import MatchCompetition
from cevlib.types.iType import IType
from cevlib.types.playByPlay import PlayByPlay
from cevlib.types.results import Result
from cevlib.types.team import Team
from cevlib.types.types import MatchState
from cevlib.types.stats import TopPlayers
from cevlib.types.report import MatchReport
from cevlib.types.info import Info


T = TypeVar("T")
FunOrProp = Union[Coroutine[Any, Any, T], T]


class IMatch(IType):
    """simple match interface"""
    @abstractmethod
    def state(self) -> FunOrProp[MatchState]:
        """state of the match"""

    @abstractmethod
    def finished(self) -> FunOrProp[bool]:
        """match finished?"""

    @abstractmethod
    def competition(self) -> FunOrProp[Optional[MatchCompetition]]:
        """competition"""

    @abstractmethod
    def homeTeam(self) -> FunOrProp[Team]:
        """home team"""

    @abstractmethod
    def awayTeam(self) -> FunOrProp[Team]:
        """away team"""

    @abstractmethod
    def venue(self) -> FunOrProp[str]:
        """venue / location"""

    @abstractmethod
    def startTime(self) -> FunOrProp[datetime]:
        """start time"""

    @abstractmethod
    def result(self) -> FunOrProp[Result]:
        """result"""

    @abstractmethod
    def matchCentreLink(self) -> Optional[str]:
        """match href"""


class IFullMatch(IMatch):
    """full match interface"""
    @abstractmethod
    def playByPlay(self) -> FunOrProp[Optional[PlayByPlay]]:
        """play by play component (match commentary)"""

    @abstractmethod
    def gallery(self) -> List[str]:
        """photo gallery"""

    @abstractmethod
    def topPlayers(self) -> FunOrProp[TopPlayers]:
        """best players"""

    @abstractmethod
    def duration(self) -> FunOrProp[timedelta]:
        """match duration"""

    @abstractmethod
    def watchLink(self) -> FunOrProp[Optional[str]]:
        """stream link"""

    @abstractmethod
    def highlightsLink(self) -> FunOrProp[Optional[str]]:
        """highlights link"""

    @abstractmethod
    def report(self) -> FunOrProp[Optional[MatchReport]]:
        """match report"""

    @abstractmethod
    def info(self) -> FunOrProp[Info]:
        """match info"""



# -*- coding: utf-8 -*-
"""cevlib"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from typing import Any, Dict, List, Optional, Type, TypeVar

T = TypeVar("T")


class DictEx(Dict[str, Any]):
    """dict extractor"""
    def __init__(self, data: Optional[Dict[str, Any]] = None) -> None:
        self._data = data or {}
        super().__init__(self._data)

    def __bool__(self) -> bool:
        return bool(self._data)

    def assertGet(self, key: str, type_: Type[T]) -> T:
        """assert has key & is type"""
        assert key in self._data
        val = self._data[key]
        assert isinstance(val, type_)
        return val

    def tryGet(self, key: str, type_: Type[T]) -> Optional[T]:
        """extracts a string"""
        try:
            return type_(self._data[key]) # type: ignore
        except: # pylint: disable=bare-except
            return None

    def ensure(self, key: str, type_: Type[T], default: Optional[T] = None) -> T:
        """get & cast w/ default"""
        try:
            return type_(self._data[key]) # type: ignore
        except: # pylint: disable=bare-except
            return default or type_()

    def ensureString(self, key: str, default: str = "") -> str:
        """extracts a string"""
        return self.ensure(key, str, default)

    def ensureInt(self, key: str, default: int = 0) -> int:
        """extracts an integer"""
        return self.ensure(key, int, default)

    def ensureBool(self, key: str, default: bool = False) -> bool:
        """extracts a boolean"""
        return self.ensure(key, bool, default)

    def ensureList(self, key: str, default: Optional[List[Any]] = None) -> List[Any]:
        """extracts a list"""
        return self.ensure(key, list, default)

    def ensureListChain(self, key: str, default: Optional[ListEx] = None) -> ListEx:
        """extracts a list and loads it in a list extractor (allowing chaining)"""
        return self.ensure(key, ListEx, default)

    def ensureDict(self, key: str, default: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        """extracts a dict"""
        return self.ensure(key, dict, default)

    def ensureDictChain(self, key: str, default: Optional[DictEx] = None) -> DictEx:
        """extracts a dict and loads it in a dict extractor (allowing chaining)"""
        return self.ensure(key, DictEx, default)


class ListEx(List[Any]):
    """list extractor"""
    def __init__(self, data: Optional[List[Any]] = None) -> None:
        self._data = data or []
        super().__init__(self._data)

    def __bool__(self) -> bool:
        return bool(self._data)

    def assertGet(self, index: int, type_: Type[T]) -> T:
        """assert has key & is type"""
        assert index < len(self._data)
        val = self._data[index]
        assert isinstance(val, type_)
        return val

    def tryGet(self, index: int, type_: Type[T]) -> Optional[T]:
        """extracts a string"""
        try:
            return type_(self._data[index]) # type: ignore
        except: # pylint: disable=bare-except
            return None

    def ensure(self, index: int, type_: Type[T], default: Optional[T] = None) -> T:
        """get & cast w/ default"""
        try:
            return type_(self._data[index]) # type: ignore
        except: # pylint: disable=bare-except
            return default or type_()

    def iterate(self, type_: Type[T], default: Optional[T] = None) -> List[T]:
        """iterate over list"""
        return [ self.ensure(i, type_, default) for i, _ in enumerate(self._data) ]

    def ensureString(self, index: int, default: str = "") -> str:
        """extracts a string"""
        return self.ensure(index, str, default)

    def ensureInt(self, index: int, default: int = 0) -> int:
        """extracts a integer"""
        return self.ensure(index, int, default)

    def ensureBool(self, index: int, default: bool = False) -> bool:
        """extracts a boolean"""
        return self.ensure(index, bool, default)

    def ensureList(self, index: int, default: Optional[List[Any]] = None) -> List[Any]:
        """extracts a list"""
        return self.ensure(index, list, default)

    def ensureListChain(self,index: int, default: Optional[ListEx] = None) -> ListEx:
        """extracts a list and loads it in a list extractor (allowing chaining)"""
        return self.ensure(index, ListEx, default)

    def ensureDict(self, index: int, default: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        """extracts a dict"""
        return self.ensure(index, dict, default)

    def ensureDictChain(self, index: int, default: Optional[DictEx] = None) -> DictEx:
        """extracts a dict and loads it in a dict extractor (allowing chaining)"""
        y = self.ensure(index, DictEx, default)
        return y



# -*- coding: utf-8 -*-
"""cevlib"""
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from cevlib.helpers.dictTool import DictEx
from cevlib.types.iType import IType, JObject


class TeamPoll(IType):
    """team poll (who will win?)"""
    def __init__(self, data: JObject) -> None:
        dex = DictEx(data)
        self._percent: float = dex.ensure("Percent", float)
        self._count: int = dex.ensure("VoteCount", int)

    def toJson(self) -> JObject:
        return {
            "percent": self.percent,
            "count": self.count
        }

    def __repr__(self) -> str:
        return f"(cevlib.types.matchPoll.TeamPoll) {self._count} ({self._percent})"

    @property
    def percent(self) -> float:
        """percentage that voted for this team (0 - 100)"""
        return self._percent

    @property
    def count(self) -> int:
        """number of votes"""
        return self._count

    @property
    def valid(self) -> bool:
        return None not in (self._percent, self._count)



# -*- coding: utf-8 -*-
"""cevlib"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

# (IMatch can be property or async)
# pylint: disable=invalid-overridden-method


import asyncio
from datetime import datetime, timedelta
import re
import json
from typing import Any, Coroutine, Dict, List, Optional, Callable
import aiohttp

from cevlib.exceptions import NotInitialisedException

from cevlib.helpers.asyncThread import asyncRunInThreadWithReturn
from cevlib.helpers.dictTool import DictEx, ListEx

from cevlib.converters.scoreHeroToJson import ScoreHeroToJson

from cevlib.types.competition import MatchCompetition
from cevlib.types.iMatch import IFullMatch
from cevlib.types.iType import JObject
from cevlib.types.info import Info
from cevlib.types.playByPlay import PlayByPlay
from cevlib.types.report import MatchReport
from cevlib.types.results import Result
from cevlib.types.stats import TeamStatistics, TopPlayer, TopPlayers
from cevlib.types.team import Team
from cevlib.types.types import MatchState


TScoreObserver = Callable[[Any, Any], Coroutine[Any, Any, Any]]


class MatchCache(IFullMatch):
    """snapshot of Match (all match data retrieved from Cache)"""
    def __init__(self,
                 playByPlay: Optional[PlayByPlay],
                 competition: Optional[MatchCompetition],
                 topPlayers: TopPlayers,
                 gallery: List[str],
                 matchCentreLink: str,
                 result: Result,
                 duration: timedelta,
                 startTime: datetime,
                 venue: str,
                 homeTeam: Team,
                 awayTeam: Team,
                 watchLink: Optional[str],
                 highlightsLink: Optional[str],
                 state: MatchState,
                 report: Optional[MatchReport],
                 info: Info) -> None:
        self._playByPlay = playByPlay
        self._competition = competition
        self._topPlayers = topPlayers
        self._gallery = gallery
        self._matchCentreLink = matchCentreLink
        self._result = result
        self._duration = duration
        self._startTime = startTime
        self._homeTeam = homeTeam
        self._awayTeam = awayTeam
        self._venue = venue
        self._watchLink = watchLink
        self._highlightsLink = highlightsLink
        self._state = state
        self._report = report
        self._info = info

    def toJson(self) -> JObject:
        return {
            "state": self.state.value,
            "result": self.result.toJson(),
            "homeTeam": self.homeTeam.toJson(),
            "awayTeam": self.awayTeam.toJson(),
            "competition": self.competition.toJson() if self.competition else None,
            "duration": str(self.duration),
            "startTime": str(self.startTime),
            "matchCentreLink": self.matchCentreLink,
            "watchLink": self.watchLink,
            "highlightsLink": self.highlightsLink,
            "venue": self.venue,
            "report": self.report.toJson() if self.report else None,
            "info": self._info.toJson() if self._info else None,
            "topPlayers": self.topPlayers.toJson(),
            "gallery": self.gallery,
            "playByPlay": self.playByPlay.toJson() if self.playByPlay else None,
        }

    @property
    def valid(self) -> bool:
        return True

    @property
    def finished(self) -> bool:
        return self.state == MatchState.Finished

    @property
    def competition(self) -> Optional[MatchCompetition]:
        return self._competition

    @property
    def playByPlay(self) -> Optional[PlayByPlay]:
        return self._playByPlay

    @property
    def topPlayers(self) -> TopPlayers:
        return self._topPlayers

    @property
    def gallery(self) -> List[str]:
        return self._gallery

    @property
    def matchCentreLink(self) -> str:
        return self._matchCentreLink

    @property
    def result(self) -> Result:
        return self._result

    @property
    def duration(self) -> timedelta:
        return self._duration

    @property
    def startTime(self) -> datetime:
        return self._startTime

    @property
    def venue(self) -> str:
        return self._venue

    @property
    def homeTeam(self) -> Team:
        return self._homeTeam

    @property
    def awayTeam(self) -> Team:
        return self._awayTeam

    @property
    def watchLink(self) -> Optional[str]:
        return self._watchLink

    @property
    def highlightsLink(self) -> Optional[str]:
        return self._highlightsLink

    @property
    def state(self) -> MatchState:
        return self._state

    @property
    def report(self) -> Optional[MatchReport]:
        return self._report

    @property
    def info(self) -> Info:
        return self._info

    @staticmethod
    async def fromMatch(match: Match) -> MatchCache:
        """create a snapshot of this match"""
        return await match.cache()

    def __repr__(self) -> str:
        return f"(cevlib.match.MatchCache) {self.toJson()}"


class Match(IFullMatch):
    """match class"""
    def __init__(self, html: str, url: str) -> None:
        self._invalidMatchCentre = "This page can be replaced with a custom 404. Check the documentation for" in html or \
                                   "Object reference not set to an instance of an object." in html # pylint: disable=line-too-long
        #if self._invalidMatchCentre:
            #raise AttributeError("404")
        self._umbracoLinks = [ match[0] for match in re.finditer(r"([\w_-]+(?:(?:\.[\w_-]+)+))([\w.,@;?^=%&:\/~+#-]*umbraco[\w.,@;?^=%&:\/~+#-]*[\w@?^=%&\/~+#-])", html) ] # pylint: disable=line-too-long
        self._gallery = [ match[0] for match in re.finditer(r"([\w_-]+(?:(?:\.[\w_-]+)+))([\w.,@;?^=%&:\/~+#-]*Upload\/Photo\/[\w .,@;?^=%&:\/~+#-]*[\w@?^=%&\/~+#-]).(jpg|JPG)", html) ] # pylint: disable=line-too-long
        embeddedVideos = [ match[0] for match in re.finditer(r"([\w_-]+(?:(?:\.[\w_-]+)+))([\w.,@;?^=%&:\/~+#-]*\/embed\/[\w .,@;?^=%&:\/~+#-]*[\w@?^=%&\/~+#-])", html) ] # pylint: disable=line-too-long
        self._highlightsLinkCache: Optional[str] = None
        if len(embeddedVideos):
            self._highlightsLinkCache = "https://" + embeddedVideos[0].replace("/embed/", "/v/") \
                                                                      .split("?")[0]
        #self._nodeId = self._getParameter(self._getLink("livescorehero"), "nodeId")
        self._html = html
        self._matchId: Optional[int] = None
        self._liveScoresCache: Optional[JObject] = None
        self._formCache: Optional[JObject] = None
        self._finished = False
        self._matchCentreLink: str = url
        self._initialised = False
        self._reportCache: Optional[MatchReport] = None
        self._infoCache: Optional[Info] = None
        self._scoreObservers: List[TScoreObserver] = [ ]
        self._scoreObserverInterval = 20
        self._init = asyncio.create_task(self._startInit())
        asyncio.create_task(self._observeScore())

    @property
    def valid(self) -> bool:
        return bool(self._umbracoLinks)

    async def _startInit(self) -> None:
        self._matchId = await self._getMatchId()
        self._initialised = True

    def init(self) -> asyncio.Task[None]:
        """
        caches the match id, required for:
        - homeTeam/awayTeam
        - currentScore
        - startTime
        - duration
        - finished
        - venue
        - watchLink
        - highlightsLink
        """
        return self._init


    # HELPERS


#    def _getParameter(self, link: str, parameter: str) -> Optional[str]:
#        try:
#            return "0"
#            # TODO working?? (not indexable)
#            #return re.search(f"(?<={parameter}=)([A-Za-z0-9]*)(?=&)?", link)[0]
#        except:
#            return None

    def _getLinks(self, contains: str) -> List[str]:
        eligibleLinks = [ ]
        for umbracoLink in self._umbracoLinks:
            if contains in umbracoLink:
                eligibleLinks.append("https://" + umbracoLink.replace("amp;", ""))
        return eligibleLinks

    def _getLink(self, contains: str, index: int = 0) -> str:
        links = self._getLinks(contains)
        if len(links) > index:
            return links[index]
        return ""

    async def _getForm(self) -> JObject:
        if self._formCache is None:
            async with aiohttp.ClientSession() as client:
                async with client.get(self._getLink("GetFormComponent")) as resp:
                    self._formCache = json.loads(await resp.json(content_type=None))
        return self._formCache

    async def _getMatchId(self) -> Optional[int]:
        try:
            async with aiohttp.ClientSession() as client:
                async with client.get(self._getLink("livescorehero")) as resp:
                    jdata = await resp.json(content_type=None)
                    return int(jdata.get("MatchId"))
        except: # pylint: disable=bare-except
            return None

    async def _requestLiveScoresJson(self, useCache: bool = True) -> JObject:
        if useCache and self._liveScoresCache:
            return self._liveScoresCache
        async with aiohttp.ClientSession() as client:
            async with client.get("https://www.cev.eu/LiveScores.json") as resp:
                jdata = DictEx(await resp.json(content_type=None))
                self._liveScoresCache = jdata
                return jdata

    async def _requestLiveScoresJsonByMatchSafe(self, useCache: bool = True) ->  Optional[JObject]:
        return await (self._requestLiveScoresJsonByMatchId(useCache) if not self._invalidMatchCentre
                      else self._requestLiveScoresJsonByMatchCentreLink(useCache))

    async def _requestLiveScoresJsonByMatchCentreLink(self,
                                                      useCache: bool = True) -> Optional[JObject]:
        assert self._matchCentreLink
        jdata = DictEx(await self._requestLiveScoresJson(useCache))
        for competition in jdata.ensure("competitions", ListEx).iterate(DictEx):
            for match in competition.ensure("matches", ListEx).iterate(DictEx):
                if match.ensure("matchCentreLink", str) == self._matchCentreLink:
                    self._finished = match.ensure("matchState_String", str) == "FINISHED"
                    match["competition"] = { "name": competition.tryGet("competitionName", str),
                                            "id": competition.tryGet("competitionId", int) }
                    return match
        return await self._tryGetFinishedGameData()

    async def _requestLiveScoresJsonByMatchId(self, useCache: bool = True) -> Optional[JObject]:
        assert self._matchId
        jdata = DictEx(await self._requestLiveScoresJson(useCache))
        for competition in jdata.ensure("competitions", ListEx).iterate(DictEx):
            for match in competition.ensure("matches", ListEx).iterate(DictEx):
                if match.ensure("matchId", int) == self._matchId:
                    self._finished = match.ensure("matchState_String", str) == "FINISHED"
                    return match
        return await self._tryGetFinishedGameData()

    async def _tryGetFinishedGameData(self, trulyFinished: bool = True) -> Optional[JObject]:
        if self._invalidMatchCentre:
            return None
        async with aiohttp.ClientSession() as client:
            self._finished = trulyFinished
            livescorehero = { }
            matchpolldata = [ ]
            async with client.get(self._getLink("getlivescorehero")) as resp:
                livescorehero = await resp.json(content_type=None)
            async with client.get(self._getLink("GetMatchPoll")) as resp:
                matchpolldata = await resp.json(content_type=None)
            return ScoreHeroToJson.convert(livescorehero, matchpolldata)

    async def _getTeam(self, index: int, home: bool) -> Optional[Team]:
        try:
            async with aiohttp.ClientSession() as client:
                playerStatsData = { }
                teamData = { }
                teamStatsData = { }
                matchPoll = [ ]
                async with client.get(self._getLink("GetStartingTeamComponent", index)) as resp:
                    teamData = await resp.json(content_type=None)
                async with client.get(self._getLink("GetPlayerStatsComponentMC")) as resp:
                    playerStatsData = json.loads(await resp.json(content_type=None))
                async with client.get(self._getLink("GetTeamStatsComponentMC")) as resp:
                    teamStatsData = json.loads(await resp.json(content_type=None))
                async with client.get(self._getLink("GetMatchPoll")) as resp:
                    matchPoll = await resp.json(content_type=None)
                liveScore = DictEx(await self._requestLiveScoresJsonByMatchSafe())
                form = await self._getForm()
                return Team(teamData,
                            playerStatsData,
                            TeamStatistics(teamStatsData, home),
                            matchPoll,
                            form["HomeTeam"] if home else form["AwayTeam"],
                            liveScore.ensure("homeTeamIcon" if home
                                             else "awayTeamIcon", str),
                            liveScore.ensure("homeTeamNickname" if home
                                             else "awayTeamNickname", str),)
        except Exception:
            liveScore = DictEx(await self._tryGetFinishedGameData(False))
            if not liveScore:
                liveScore = DictEx(await self._requestLiveScoresJsonByMatchSafe())
            return Team.build(liveScore.ensure("homeTeam" if home else "awayTeam",
                                               str),
                              liveScore.ensure("homeTeamIcon" if home else "awayTeamIcon",
                                               str),
                              liveScore.ensure("homeTeamNickname" if home else "awayTeamNickname",
                                               str),
                              home)


    # GET


    def setScoreObserverInterval(self, intervalS: int) -> None:
        """the interval the score observer uses"""
        self._scoreObserverInterval = intervalS

    def addScoreObserver(self, observer: TScoreObserver) -> None:
        """adds a new score observer"""
        self._scoreObservers.append(observer)

    def removeScoreObserver(self, observer: TScoreObserver) -> None:
        """removes a score observer"""
        self._scoreObservers.remove(observer)

    async def _observeScore(self) -> None:
        lastScore: Optional[Result] = None
        while True:
            try:
                if len(self._scoreObservers) > 0:
                    result = await self.result()
                    if lastScore != result:
                        lastScore = result
                        for observer in self._scoreObservers:
                            await observer(self, lastScore)
            except: # pylint: disable=bare-except
                pass
            await asyncio.sleep(self._scoreObserverInterval)

    async def result(self) -> Result:
        if not self._initialised:
            raise NotInitialisedException
        match = await self._requestLiveScoresJsonByMatchSafe(self._finished)
        assert match is not None
        return Result(match)

    async def startTime(self) -> datetime:
        if not self._initialised:
            raise NotInitialisedException
        match = DictEx(await self._requestLiveScoresJsonByMatchSafe())
        return datetime.strptime(match.ensure("utcStartDate", str), "%Y-%m-%dT%H:%M:%SZ")

    async def _started(self) -> bool:
        startTime = await self.startTime()
        return datetime.utcnow() >= startTime

    async def _finishedF(self) -> bool:
        if not self._initialised:
            await self.init()
        await self._requestLiveScoresJsonByMatchSafe(False)
        return self._finished

    async def finished(self) -> bool:
        return await self._finishedF()

    async def state(self) -> MatchState:
        started, finished = await asyncio.gather(self._started(), self._finishedF())
        return MatchState.parse(started, finished)

    async def venue(self) -> str:
        if not self._initialised:
            raise NotInitialisedException
        match = DictEx(await self._requestLiveScoresJsonByMatchSafe())
        return match.ensure("matchLocation", str)

    async def playByPlay(self) -> Optional[PlayByPlay]:
        try:
            async with aiohttp.ClientSession() as client:
                async with client.get(self._getLink("GetPlayByPlayComponent")) as resp:
                    jdata = await resp.json(content_type=None)
                    return PlayByPlay(jdata)
        except Exception:
            return None

    async def homeTeam(self) -> Team:
        if not self._initialised:
            raise NotInitialisedException
        team = await self._getTeam(0, True)
        assert team
        return team

    async def awayTeam(self) -> Team:
        if not self._initialised:
            raise NotInitialisedException
        team = await self._getTeam(1, False)
        assert team
        return team

    @property
    def gallery(self) -> List[str]:
        return self._gallery

    async def report(self) -> Optional[MatchReport]:
        if not self._reportCache:
            self._reportCache = await asyncRunInThreadWithReturn(MatchReport, self._html)
        return self._reportCache or None

    async def duration(self) -> timedelta:
        if not self._initialised:
            raise NotInitialisedException
        if not await self._finishedF():
            startTime = await self.startTime()
            if datetime.utcnow() < startTime:
                return timedelta()
            return datetime.utcnow() - startTime
        if self._invalidMatchCentre:
            return timedelta()
        async with aiohttp.ClientSession() as client:
            async with client.get(self._getLink("getlivescorehero")) as resp:
                jdata = await resp.json(content_type=None)
                return timedelta(minutes = float(jdata.get("Duration").split(" ")[0]))

    @property
    def matchCentreLink(self) -> str:
        return self._matchCentreLink

    async def watchLink(self) -> Optional[str]:
        if not self._initialised:
            raise NotInitialisedException
        jdata = DictEx(await self._requestLiveScoresJsonByMatchSafe())
        link = jdata.tryGet("watchLink", str)
        if link:
            return link

        async with aiohttp.ClientSession() as client:
            async with client.get(self._getLink("getlivescorehero")) as resp:
                jdata = DictEx(await resp.json(content_type=None))
                return jdata.tryGet("HighLightUrl", str)

    async def highlightsLink(self) -> Optional[str]:
        if not self._highlightsLinkCache:
            if not self._initialised:
                raise NotInitialisedException
            jdata = DictEx(await self._requestLiveScoresJsonByMatchSafe())
            self._highlightsLinkCache = jdata.tryGet("highlightsLink", str)

            if not self._highlightsLinkCache:
                async with aiohttp.ClientSession() as client:
                    async with client.get(self._getLink("getlivescorehero")) as resp:
                        jdata = DictEx(await resp.json(content_type=None))
                        self._highlightsLinkCache = jdata.tryGet("HighLightUrl", str)

        return self._highlightsLinkCache

    async def competition(self) -> Optional[MatchCompetition]:
        if self._invalidMatchCentre:
            jdata = DictEx(await self._requestLiveScoresJsonByMatchSafe())
            competition = jdata.ensure("competition", DictEx)
            if not competition:
                return None
            return MatchCompetition({
                "Competition": competition.ensure("name", str),
                "Leg": jdata.ensure("legName", str),
                "Phase": jdata.ensure("phaseName", str),
                "GroupPool": jdata.ensure("groupName", str),
                "MatchNumber": jdata.ensure("matchNumber", str)
            })
        async with aiohttp.ClientSession() as client:
            async with client.get(self._getLink("getlivescorehero")) as resp:
                jdata = await resp.json(content_type=None)
                assert jdata is not None
                return MatchCompetition(jdata)

    async def topPlayers(self) -> TopPlayers:
        topPlayers = TopPlayers()
        links = self._getLinks("GetTopStatisticsComponent")
        for link in links:
            async with aiohttp.ClientSession() as client:
                async with client.get(link) as resp:
                    jdata = await resp.json(content_type=None)
                    topPlayers.append(TopPlayer(jdata))
        return topPlayers

    async def info(self) -> Info:
        if self._infoCache is None:
            self._infoCache = await asyncRunInThreadWithReturn(Info, self._html)
        assert self._infoCache
        return self._infoCache


    # CREATE


    @staticmethod
    async def byUrl(url: str) -> Match:
        """creates a match by match url (link/href)"""
        async with aiohttp.ClientSession() as client:
            async with client.get(url) as resp:
                return Match(await resp.text(), url)


    # CONVERT


    async def toJson(self) -> Dict[str, Any]:
        return (await self.cache()).toJson()

    async def cache(self,) -> MatchCache:
        """
        gets a snapshot of the current data.
        takes about 1.2s and is thus significantly faster than
        direct queries of the properties
        """
                    # might allow selecting properties (in a future version)
                    #playByplay = True,
                    #competition = True,
                    #topPlayers = True,
                    #currentScore = True,
                    #duration = True,
                    #startTime = True,
                    #venue = True,
                    #homeTeam = True,
                    #awayTeam = True,
                    #watchLink = True,
                    #highlightsLink = True,
                    #finished = True,
                    #report = True,) -> MatchCache:
        afterInit: List[Coroutine[Any, Any, Any]] = [ ]

        afterInit.append(self.playByPlay())
        afterInit.append(self.competition())
        afterInit.append(self.topPlayers())
        afterInit.append(self.report())
        afterInit.append(self.info())

        if not self._initialised:
            await self.init()

        afterInit.append(self.result())
        afterInit.append(self.duration())
        afterInit.append(self.startTime())
        afterInit.append(self.venue())
        afterInit.append(self.homeTeam())
        afterInit.append(self.awayTeam())
        afterInit.append(self.watchLink())
        afterInit.append(self.highlightsLink())
        afterInit.append(self.state())
        afterInitResults = await asyncio.gather(*afterInit)

        return MatchCache(playByPlay= afterInitResults[0],
                          competition= afterInitResults[1],
                          topPlayers= afterInitResults[2],
                          gallery= self.gallery,
                          matchCentreLink= self._matchCentreLink,
                          result= afterInitResults[5],
                          duration= afterInitResults[6],
                          startTime= afterInitResults[7],
                          venue= afterInitResults[8],
                          homeTeam= afterInitResults[9],
                          awayTeam= afterInitResults[10],
                          watchLink= afterInitResults[11],
                          highlightsLink= afterInitResults[12],
                          state= afterInitResults[13],
                          report= afterInitResults[3],
                          info= afterInitResults[4])
